#include "Context.h"
#include "Constant.h"
#include "PetscWrapper.h"
#include "Utils.h"


using namespace Field2D;


namespace
{
	const PetscInt X_ = 0;
	const PetscInt Y_ = 1;
}

PetscErrorCode Context::CreateDMDA
(
	Ctx &ctx
)
{
	DMDACreate2d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		(Constant::Grid::Nx), //global dimension in x-direction of the array
		(Constant::Grid::Ny), //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		2,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		&ctx.da               //dynamic array
	);
	DMSetUp(ctx.da);
	DMDASetUniformCoordinates
	(
		ctx.da,
		0.,(Constant::Grid::L[0]),
		0.,(Constant::Grid::L[1]),
		0.,(Constant::Grid::L[2])
	);
	return(0);
}


PetscErrorCode Context::DestroyDMDA
(
	Ctx &ctx
)
{
	DMDestroy(&ctx.da);
	return(0);
}


PetscErrorCode Context::CreateFields
(
	Ctx &ctx,
	const PetscInt n
)
{
	//1D fields
	CreateField(ctx.da,ctx.phiSum,"phiSum");
	CreateField(ctx.da,ctx.phiWall,"phiWall"); if (Constant::Foam::ImportWall){ReadFieldInBinaryFormat(ctx.phiWall);}
	CreateField(ctx.da,ctx.rho,"rho"); SetField(ctx.rho,1);
	CreateField(ctx.da,ctx.eta,"eta"); SetField(ctx.eta,1);
	CreateField(ctx.da,ctx.p,"p");
	CreateField(ctx.da,ctx.divRHS,"divRHS");

	//2D fields
	Create2DField(ctx.da,ctx.u,"u");
	Create2DField(ctx.da,ctx.uRHS,"uRHS");
	Create2DField(ctx.da,ctx.fConv,"fConv");
	Create2DField(ctx.da,ctx.fVisc,"fVisc");
	Create2DField(ctx.da,ctx.fViscT,"fViscT");
	Create2DField(ctx.da,ctx.fCap,"fCap");
	Create2DField(ctx.da,ctx.fExt,"fExt");
	Create2DField(ctx.da,ctx.nv,"nv");
	Create2DField(ctx.da,ctx.sFTmp,"sFTmp");

	//Foam-associated fields
	CreateField(ctx.da,ctx.phi,"phi",n);
	CreateField(ctx.da,ctx.phiRHS,"phiRHS",n);
	CreateField(ctx.da,ctx.mu,"mu",n);
	CreateField(ctx.da,ctx.muBulk,"muBulk",n);
	CreateField(ctx.da,ctx.muRep,"muRep",n);
	CreateField(ctx.da,ctx.sFx,"sFx",n);
	CreateField(ctx.da,ctx.sFy,"sFy",n);
	CreateField(ctx.da,ctx.pGas,"pGas",n);
	return(0);
}


PetscErrorCode Context::DestroyFields
(
	Ctx &ctx
)
{
	//DA
	DMDestroy(&ctx.da);

	//1D fields
	DestroyField(ctx.phiSum);
	DestroyField(ctx.phiWall);
	DestroyField(ctx.rho);
	DestroyField(ctx.eta);
	DestroyField(ctx.p);
	DestroyField(ctx.divRHS);

	//2D fields
	DestroyField(ctx.u);
	DestroyField(ctx.uRHS);
	DestroyField(ctx.fConv);
	DestroyField(ctx.fVisc);
	DestroyField(ctx.fViscT);
	DestroyField(ctx.fCap);
	DestroyField(ctx.fExt);
	DestroyField(ctx.nv);
	DestroyField(ctx.sFTmp);

	//Foam-associated fields
	DestroyField(ctx.phi);
	DestroyField(ctx.phiRHS);
	DestroyField(ctx.mu);
	DestroyField(ctx.muBulk);
	DestroyField(ctx.muRep);
	DestroyField(ctx.sFx);
	DestroyField(ctx.sFy);
	DestroyField(ctx.pGas);

	return(0);
}


PetscErrorCode Context::CreateOneDimensionalArrays
(
	Ctx &ctx,
	const PetscInt n
)
{
	PetscMalloc1(n,&ctx.bubbleVolume);
	PetscMalloc1(n,&ctx.initialBubbleVolume);
	return(0);
}


PetscErrorCode Context::DestroyOneDimensionalArrays
(
	Ctx &ctx
)
{
	PetscFree(ctx.bubbleVolume);
	PetscFree(ctx.initialBubbleVolume);
	return(0);
}


PetscErrorCode Context::SetScalars
(
	Ctx &ctx
)
{
	PetscScalar invPe = 1./(Constant::Fluid::Pe);
	PetscScalar invRe = 1./(Constant::Fluid::Re);
	PetscScalar invWe = 1./(Constant::Fluid::We);
	ctx.dt = (Constant::Time::DeltaT);
	ctx.dx = (Constant::Grid::Dx);
	ctx.invDx1 = 1./(ctx.dx);
	ctx.invDx2 = (ctx.invDx1)*(ctx.invDx1);
	ctx.invDx3 = (ctx.invDx1)*(ctx.invDx2);
	ctx.xi = Constant::Fluid::Xi;
	ctx.xi2 = PetscSqr(ctx.xi);
	ctx.dtInvDx1 = (ctx.dt ) * (ctx.invDx1);
	ctx.dtInvPe = (ctx.dt ) * invPe;
	ctx.xi2DtInvPeExp = (ctx.xi2) * (ctx.dt) * invPe * (1.- (Constant::Numerical::ThetaPhi) );
	ctx.xi2DtInvPeImp = (ctx.xi2) * (ctx.dt) * invPe * (    (Constant::Numerical::ThetaPhi) );
	ctx.dtInvDx2InvPe = (ctx.dt) * (ctx.invDx2) * invPe;
	ctx.dtInvDx2InvRe = (ctx.dt) * (ctx.invDx2) * invRe;
	ctx.dtXiInvDx3InvWe = (3. * PetscSqrtScalar(2.) ) * (ctx.dt) * (ctx.xi) * (ctx.invDx3) * invWe;
	ctx.bubbleGrowthCoefficient = (Constant::Foam::BubbleGrowthCoef);
	ctx.repulsionCoefficient = (Constant::Foam::BubbleRepulsionCoef);
	return(0);
}


PetscErrorCode Context::CreateContext
(
	Ctx &ctx,
	const PetscInt n
)
{
	Context::CreateDMDA(ctx);
	Context::CreateFields(ctx,n);
	Context::CreateOneDimensionalArrays(ctx,n);
	Context::SetScalars(ctx);
	ctx.tSave = 0;
	PetscPrintf(PETSC_COMM_WORLD,"  Solution context created\n");
	return(0);
}


PetscErrorCode Context::DestroyContext
(
	Ctx &ctx
)
{
	Context::DestroyDMDA(ctx);
	Context::DestroyFields(ctx);
	Context::DestroyOneDimensionalArrays(ctx);
	return(0);
}


PetscErrorCode Context::WriteContext
(
	Ctx &ctx
)
{
	WriteFieldInBinaryFormat(ctx.phi,ctx.tSave);
	WriteFieldInVTSFormat(ctx.phiSum,ctx.tSave);
//DEBUG
// 	WriteFieldInVTSFormat(ctx.u,ctx.tSave);
// 	WriteFieldInVTSFormat(ctx.eta,ctx.tSave);
// 	WriteFieldInVTSFormat(ctx.rho,ctx.tSave);
// 	WriteFieldInVTSFormat(ctx.uRHS,ctx.tSave);
// 	WriteFieldInVTSFormat(ctx.divRHS,ctx.tSave);
// 	WriteFieldInVTSFormat(ctx.p,ctx.tSave);
// 	WriteFieldInVTSFormat(ctx.fCap,ctx.tSave);
// 	WriteFieldInVTSFormat(ctx.sFx,ctx.tSave);
//DEBUG
	if ( ctx.tSave < 1 )
	{
		WriteFieldInBinaryFormat(ctx.phiWall);
		WriteFieldInVTSFormat(ctx.phiWall);
		PetscWrapper::WriteScalarInBinaryFormat(ctx.initialBubbleVolume,"initialBubbleVolume",ctx.phi.n);
		PetscWrapper::WriteScalarInBinaryFormat(ctx.phi.n,"bubbleNumber");
	}
	ctx.tSave++;
	return(0);
}


PetscErrorCode Context::ReadContext
(
	Ctx &ctx,
	const PetscInt t
)
{
	ReadFieldInBinaryFormat(ctx.phi,t);
	PetscWrapper::ReadScalarInBinaryFormat(ctx.initialBubbleVolume,"initialBubbleVolume",ctx.phi.n);
	Context::SetScalars(ctx);
	ctx.tSave = t+1;
	ForAllB(ctx.phi.n)
	{
		PetscPrintf(PETSC_COMM_WORLD,"  V[%i] = %4.2f \n",b,ctx.initialBubbleVolume[b]);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Solution context read from save = %i\n",t);
	return(0);
}


PetscErrorCode Context::InitFoamToSingleBubble
(
	Ctx &ctx
)
{
	const PetscScalar Lx = (Constant::Grid::L[X_]);
	const PetscScalar Ly = (Constant::Grid::L[Y_]);
	const PetscScalar HLx = (Constant::Grid::HL[X_]);
	const PetscScalar HLy = (Constant::Grid::HL[Y_]);
	const PetscScalar Xb = HLx;
	const PetscScalar Yb = HLy;
	const PetscScalar Rb = Ly/6;
	const PetscScalar xi = PetscSqrtScalar(2.) * (Constant::Fluid::Xi);
// 	const PetscScalar xi = 0.0001;
	PetscScalar lV = 0;
	PetscScalar gV;
	PetscWrapper::GridCoordinates gridCoordinates;
	PetscWrapper::GetPetscArray(ctx.da,gridCoordinates);
	GetFieldArray(ctx.phi);
	ForAllJI(ctx.da)
	{
		PetscScalar rx = Xb - gridCoordinates.x(j,i); if (rx <= -HLx){rx += Lx;} if (rx >= HLx){rx -= Lx;}
		PetscScalar ry = Yb - gridCoordinates.y(j,i); if (ry <= -HLy){ry += Ly;} if (ry >= HLy){ry -= Ly;}
		PetscScalar r = PetscSqrtScalar(rx*rx+ry*ry);
		PetscScalar localPhi = 0.5 + 0.5 * PetscTanhScalar( (Rb-r) / xi );
		ctx.phi(0,j,i) = localPhi;
		lV += localPhi;
	}
	PetscWrapper::RestorePetscArray(ctx.da,gridCoordinates);
	RestoreFieldArray(ctx.phi);
	MPI_Allreduce(&lV,&gV,1,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
	ctx.initialBubbleVolume[0] = gV;
	PetscPrintf(PETSC_COMM_WORLD,"  Single bubble created\n");
	return(0);
}


// 	ReadFieldInBinaryFormat(ctx.u,t);
// 	if (Constant::Foam::VZero < 0)
// 	{
// 		PetscWrapper::ReadScalarInBinaryFormat(ctx.initialBubbleVolume,"initialBubbleVolume",ctx.phi.n);
// 	}
// 	else
// 	{
// 		PetscPrintf(PETSC_COMM_WORLD,"  Set bubble volume\n");
// 		ForAllB(ctx.phi.n)
// 		{
// 			ctx.initialBubbleVolume[b] = Constant::Foam::VZero;
// 		}
// 		PetscWrapper::WriteScalarInBinaryFormat(ctx.initialBubbleVolume,"initialBubbleVolume",ctx.phi.n);
// 	}
