#ifndef CONTEXT_H
#define CONTEXT_H


#include "Field2D.h"


/** \file
 *  \brief Solution context */


/** \namespace Context
 *  \brief Solution context */
namespace Context
{


	/** \brief Solution context */
	class Ctx
	{
	public:

		//Petsc dynamic array
		DM da;

		//One-dimensional fields
		Field2D::ScalarField phiSum; /**< \brief Sum of all phi */
		Field2D::ScalarField phiWall; /**< \brief Smooth Wall */
		Field2D::ScalarField rho; /**< \brief Density */
		Field2D::ScalarField eta; /**< \brief Viscosity */
		Field2D::ScalarField p; /**< \brief Pressure */
		Field2D::ScalarField divRHS; /**< \brief div(RHS) */

		//Two-dimensional fields
		Field2D::VectorField u; /**< \brief Velocity */
		Field2D::VectorField uRHS; /**< \brief Right hand side velocity */
		Field2D::VectorField fConv; /**< \brief convective acceleration */
		Field2D::VectorField fVisc; /**< \brief Viscous acceleration */
		Field2D::VectorField fViscT; /**< \brief Transposed viscous acceleration */
		Field2D::VectorField fCap; /**< \brief Capillary acceleration */
		Field2D::VectorField fExt; /**< \brief External acceleration */
		Field2D::VectorField nv; /**< \brief Temporary normal vector used in capillary force calculation \f$ \nabla \phi_b \f$ */
		Field2D::VectorField sFTmp; /**< \brief Temporary surface force used in capillary force calculation */

		//N-dimensional fields
		Field2D::VectorField phi;  /**< \brief Foam */
		Field2D::VectorField phiRHS; /**< \brief Phi right hand side */
		Field2D::VectorField mu; /**< \brief Chemical potential */
		Field2D::VectorField muBulk; /**< \brief Bulk part of the chemical potential */
		Field2D::VectorField muRep; /**< \brief Repulsive part of the chemical potential */
		Field2D::VectorField sFx; /**< \brief X-component of surface force */
		Field2D::VectorField sFy; /**< \brief Y-component of surface force */
		Field2D::VectorField pGas; /**< \brief Pressure for each gas bubble */

		//Arrays
		PetscScalar *bubbleVolume; /**< \brief Array of bubble volumes */
		PetscScalar *initialBubbleVolume; /**< \brief Array of initial bubble volumes */

		//Integers
		PetscInt b; /**< \brief Bubble index needed in the SNES solver */
		PetscInt tSave; /**< \brief Incremental save index */

		//Scalars
		PetscScalar dt; /**< \brief time step */
		PetscScalar dx; /**< \brief dx */
		PetscScalar invDx1; /**< \brief 1 / dx */
		PetscScalar invDx2; /**< \brief 1 / (dx^2) */
		PetscScalar invDx3; /**< \brief 1 / (dx^3) */
		PetscScalar xi; /**< \brief xi */
		PetscScalar xi2; /**< \brief xi^2 */
		PetscScalar dtInvDx1; /**< \brief -dt/dx */
		PetscScalar dtInvPe; /**< \brief -dt/Pe */
		PetscScalar xi2DtInvPeExp; /**< \brief xi^2 . dt / Pe (Explicit term) */
		PetscScalar xi2DtInvPeImp; /**< \brief xi^2 . dt / Pe (Implicit term) */
		PetscScalar dtInvDx2InvPe; /**< \brief dt / (Pe . dx^2) */
		PetscScalar dtInvDx2InvRe; /**< \brief dt / (Re . dx^2) */
		PetscScalar dtXiInvDx3InvWe; /**< \brief dt . xi / (We . dx^3) */

		PetscScalar bubbleGrowthCoefficient; /**< \brief Bubble growth coefficient */
 		PetscScalar repulsionCoefficient; /**< \brief Repulsion coefficient */
 		PetscScalar avgRho; /**< \brief Average density */
	};


	/** \brief Set scalars present in the solution context
	 *  \param[in] ctx solution context */
	PetscErrorCode SetScalars(Ctx &ctx);


	/** \brief Create DMDA dynamic array
	 *  \param[in] ctx solution context */
	PetscErrorCode CreateDMDA(Ctx &ctx);


	/** \brief Destroy DMDA dynamic array
	 *  \param[in] ctx solution context */
	PetscErrorCode DestroyDMDA(Ctx &ctx);


	/** \brief Create fields
	 *  \param[in] ctx solution context
	    \param[in] n   number of bubbles */
	PetscErrorCode CreateFields(Ctx &ctx,const PetscInt n);


	/** \brief Destroy fields
	 *  \param[in] ctx solution context */
	PetscErrorCode DestroyFields(Ctx &ctx);


	/** \brief Create one-dimensional arrays
	 *  \param[in] ctx solution context
	    \param[in] n   number of bubbles */
	PetscErrorCode CreateOneDimensionalArrays(Ctx &ctx,const PetscInt n);


	/** \brief Destroy one-dimensional arrays
	 *  \param[in] ctx solution context */
	PetscErrorCode DestroyOneDimensionalArrays(Ctx &ctx);


	/** \brief Create context
	 *  \param[in] ctx solution context
	    \param[in] n   number of bubbles */
	PetscErrorCode CreateContext(Ctx &ctx,const PetscInt n);


	/** \brief Destroy context
	 *  \param[in] ctx solution context */
	PetscErrorCode DestroyContext(Ctx &ctx);


	/** \brief Read context
	 *  \param[in] ctx solution context
	 *  \param[in] t   time index */
	PetscErrorCode ReadContext(Ctx &ctx,const PetscInt t);


	/** \brief Write context
	 *  \param[in] ctx solution context */
	PetscErrorCode WriteContext(Ctx &ctx);


	/** \brief Create one single bubble at channel centre
	 *  \param[in] ctx solution context */
	PetscErrorCode InitFoamToSingleBubble(Ctx &ctx);


}


#endif
