#include "petsc.h"
#include "Constant.h"
#include "PetscWrapper.h"
#include "Field2D.h"
#include "Utils.h"

static char help[] = "Create wall as smooth field.\n\n";
const PetscInt X_ = 0;
const PetscInt Y_ = 1;

PetscErrorCode CreateDMDA
(
	DM &da
)
{
	DMDACreate2d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		(Constant::Grid::Nx), //global dimension in x-direction of the array
		(Constant::Grid::Ny), //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		2,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		&da                //dynamic array
	);
	DMSetUp(da);
	DMDASetUniformCoordinates
	(
		da,
		0.,(Constant::Grid::L[0]),
		0.,(Constant::Grid::L[1]),
		0.,(Constant::Grid::L[2])
	);
	return(0);
}


PetscErrorCode DestroyDMDA
(
	DM &da
)
{
	DMDestroy(&da);
	return(0);
}


PetscInt main( int argc, char* argv[] )
{
	PetscInitialize(&argc,&argv,(char*)0,help);

	//Constant
	Constant::ReadInputFile();

	//Create
	DM da;
	CreateDMDA(da);
	Field2D::ScalarField phiWall;
	Field2D::CreateField(da,phiWall,"phiWall");

	//Coordinates
	PetscWrapper::GridCoordinates gridCoordinates;

	//Domain dimensions
	const PetscScalar Lx = (Constant::Grid:: L[X_]);
	const PetscScalar Ly = (Constant::Grid:: L[Y_]);
	const PetscScalar HLx = (Constant::Grid::HL[X_]);
	const PetscScalar HLy = (Constant::Grid::HL[Y_]);
	const PetscScalar dx = (Constant::Grid::Dx);

	//Radius
	const PetscScalar Rf = 1;

	//Diffuse interfacial wall length
	const PetscScalar xi  = PetscSqrtScalar(2.) * (Constant::Fluid::Xi);

	//Wall
	PetscWrapper::GetPetscArray(da,gridCoordinates);
	Field2D::GetFieldArray(phiWall);
	PetscScalar a  = 0.002;
	PetscScalar b  =-0.45;
	PetscScalar c  = 33.3;
	PetscScalar d  =-750;
	PetscScalar x0 = 50;
	PetscScalar x1 = 100;
	ForAllJI(da)
	{
		PetscInt Nf = 100;
		PetscScalar sum = 0;
		for (PetscInt s=0; s<Nf; ++s)
		{
			PetscScalar nomi =  s;
			PetscScalar denomi = Nf-1;
			PetscScalar alpha =  nomi / denomi;
			PetscScalar xc = x0 + alpha * (x1-x0);
			PetscScalar yc = a*(xc*xc*xc) + b*(xc*xc) + c*(xc) + d;
// 			PetscPrintf(PETSC_COMM_WORLD,"%4.2f (%4.2f,%4.2f)\n",alpha,xc,yc);
			PetscScalar rx = xc-gridCoordinates.x(j,i);
			PetscScalar ry = yc-gridCoordinates.y(j,i);
			PetscScalar r  = PetscSqrtScalar(rx*rx+ry*ry);
			PetscScalar phi = r<Rf ? 1 : 0;
			phi = 0.5 + 0.5 * PetscTanhScalar( (Rf-r) / xi );
			sum += phi;
		}

		//Add
		phiWall(j,i) = sum/PetscMax(sum,1);
	}
	PetscWrapper::RestorePetscArray(da,gridCoordinates);
	Field2D::RestoreFieldArray(phiWall);

	//Write
	Field2D::WriteFieldInBinaryFormat(phiWall);
	Field2D::WriteFieldInVTSFormat(phiWall);

	//Destroy
	Field2D::DestroyField(phiWall);
	DestroyDMDA(da);
	PetscFinalize();
	return(0);
}
