#include "petsc.h"
#include "Constant.h"
#include "PetscWrapper.h"
#include "Field2D.h"
#include "Utils.h"

static char help[] = "Create foam.\n\n";
const PetscInt X_ = 0;
const PetscInt Y_ = 1;


PetscErrorCode CreateDMDA
(
	DM &da
)
{
	DMDACreate2d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		(Constant::Grid::Nx), //global dimension in x-direction of the array
		(Constant::Grid::Ny), //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		2,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		&da                   //dynamic array
	);
	DMSetUp(da);
	DMDASetUniformCoordinates
	(
		da,
		0.,(Constant::Grid::L[0]),
		0.,(Constant::Grid::L[1]),
		0.,(Constant::Grid::L[2])
	);
	return(0);
}


PetscErrorCode DestroyDMDA
(
	DM &da
)
{
	DMDestroy(&da);
	return(0);
}


PetscInt main( int argc, char* argv[] )
{
	PetscInitialize(&argc,&argv,(char*)0,help);

	//Constant
	Constant::ReadInputFile();

	//Disk radius
	PetscScalar rd = 0.2*(Constant::Grid::L[Y_]);
	PetscOptionsGetReal(NULL,NULL,"-disk_radius",&rd,NULL);

	//Gap
	PetscScalar gap = 0.01*(Constant::Grid::L[Y_]);
	PetscOptionsGetReal(NULL,NULL,"-gap",&gap,NULL);

	//Napprox
	PetscInt napprox = 90;
	PetscOptionsGetInt(NULL,NULL,"-napprox",&napprox,NULL);

	//Get exact number of bubbles in x and y directions from estimate
	PetscScalar Lx = (Constant::Grid::L[X_]);
	PetscScalar Ly = (Constant::Grid::L[Y_]) - 2.*rd;
	PetscScalar alpha = PetscMax(Ly/Lx, Lx/Ly);
	PetscInt tmpx = PetscCeilReal( PetscSqrtReal( static_cast<PetscScalar>(napprox) * alpha) );
	PetscInt tmpy = PetscCeilReal( static_cast<PetscScalar>(napprox) / static_cast<PetscScalar>(tmpx) ) ;
	PetscInt nx = Lx > Ly ? tmpx:tmpy;
	PetscInt ny = Lx > Ly ? tmpy:tmpx;
	PetscInt n = nx*ny;

	//Print n
	PetscPrintf(PETSC_COMM_WORLD,"  Nb = %i\n",n);

	//Partition domain into rectangles
	PetscScalar dx = Lx/static_cast<PetscScalar>(nx);
	PetscScalar dy = Ly/static_cast<PetscScalar>(ny);

	//Create
	DM da;
	CreateDMDA(da);
	Field2D::VectorField phi;
	Field2D::ScalarField phiSum;
	Field2D::CreateField(da,phi,"phi",n);
	Field2D::CreateField(da,phiSum,"phiSum");

	//Coordinates
	PetscWrapper::GridCoordinates gridCoordinates;

	//Foam
	PetscWrapper::GetPetscArray(da,gridCoordinates);
	Field2D::GetFieldArray(phi);
	for (PetscInt p=0; p<ny; p++)
	{
		for (PetscInt q=0; q<nx; q++)
		{
			//Get bubble index b = 0..n-1
			PetscInt b = p*nx + q;

			//Partition domain into rectangles
			PetscScalar ymin = static_cast<PetscScalar>(p  ) * dy + gap + rd;
			PetscScalar ymax = static_cast<PetscScalar>(p+1) * dy - gap + rd;
			PetscScalar xmin = static_cast<PetscScalar>(q  ) * dx + gap ;
			PetscScalar xmax = static_cast<PetscScalar>(q+1) * dx - gap ;
			ForAllJI(da)
			{
				//Grid point coordinate
				const PetscScalar& xc = gridCoordinates.x(j,i);
				const PetscScalar& yc = gridCoordinates.y(j,i);

				//Set phi = 1 inside b-th bubble
				if ( (xmin < xc && xc < xmax) && (ymin < yc && yc < ymax) )
				{
					phi(b,j,i) = 1.00;
				}
			}
		}
	}
	PetscWrapper::RestorePetscArray(da,gridCoordinates);
	Field2D::RestoreFieldArray(phi);

	//Write
	Field2D::SumVectorField(phi,phiSum);
	Field2D::WriteFieldInBinaryFormat(phi,0);
	Field2D::WriteFieldInVTSFormat(phiSum,0);
	PetscWrapper::WriteScalarInBinaryFormat(n,"bubbleNumber");

	//Destroy
	Field2D::DestroyField(phi);
	Field2D::DestroyField(phiSum);
	DestroyDMDA(da);
	PetscFinalize();
	return(0);
}
