#include "petsc.h"
#include "Constant.h"
#include "PetscWrapper.h"
#include "Field2D.h"
#include "Utils.h"

static char help[] = "Create wall as smooth field.\n\n";
const PetscInt X_ = 0;
const PetscInt Y_ = 1;

PetscErrorCode CreateDMDA
(
	DM &da
)
{
	DMDACreate2d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		(Constant::Grid::Nx), //global dimension in x-direction of the array
		(Constant::Grid::Ny), //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		2,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		&da                //dynamic array
	);
	DMSetUp(da);
	DMDASetUniformCoordinates
	(
		da,
		0.,(Constant::Grid::L[0]),
		0.,(Constant::Grid::L[1]),
		0.,(Constant::Grid::L[2])
	);
	return(0);
}


PetscErrorCode DestroyDMDA
(
	DM &da
)
{
	DMDestroy(&da);
	return(0);
}


PetscInt main( int argc, char* argv[] )
{
	PetscInitialize(&argc,&argv,(char*)0,help);

	//Constant
	Constant::ReadInputFile();

	//Create
	DM da;
	CreateDMDA(da);
	Field2D::ScalarField phiWall;
	Field2D::CreateField(da,phiWall,"phiWall");

	//Coordinates
	PetscWrapper::GridCoordinates gridCoordinates;

	//Domain dimensions
	const PetscScalar Lx = (Constant::Grid:: L[X_]);
	const PetscScalar Ly = (Constant::Grid:: L[Y_]);
	const PetscScalar HLx = (Constant::Grid::HL[X_]);
	const PetscScalar HLy = (Constant::Grid::HL[Y_]);
	const PetscScalar dx = (Constant::Grid::Dx);

	//Band dimensions
	const PetscScalar Yb = 0.;    //Band centre
	const PetscScalar Rb = Ly/30; //Half band width (5*dx with Ly = 150 dx)

	//Disk dimensions
	const PetscScalar Xd = HLx;     //X-centre
	const PetscScalar Yd = Rb;      //Y-centre
	const PetscScalar Rd = 0.18*Ly; //Radius = 0.22

	//Diffuse interfacial wall length
	const PetscScalar xi  = PetscSqrtScalar(2.) * (Constant::Fluid::Xi);

	//Wall
	PetscWrapper::GetPetscArray(da,gridCoordinates);
	Field2D::GetFieldArray(phiWall);
	ForAllJI(da)
	{
		//Band
		PetscScalar phi_b;
		{
			PetscScalar rx = 0.0;                       if (rx <= -HLx){rx += Lx;} if (rx >= HLx){rx -= Lx;}
			PetscScalar ry = Yb-gridCoordinates.y(j,i); if (ry <= -HLy){ry += Ly;} if (ry >= HLy){ry -= Ly;}
			PetscScalar r = PetscSqrtScalar(rx*rx+ry*ry);
			phi_b = 0.5 + 0.5 * PetscTanhScalar( (Rb-r) / xi );
		}

		//Lower disc
		PetscScalar phi_ld;
		{
			PetscScalar rx = Xd-gridCoordinates.x(j,i);
			PetscScalar ry = Yd-gridCoordinates.y(j,i);
			PetscScalar r = PetscSqrtScalar(rx*rx+ry*ry);
			phi_ld = 0.5 + 0.5 * PetscTanhScalar( (Rd-r) / xi );
		}

		//Upper disc
		PetscScalar phi_ud;
		{
			PetscScalar rx = (Xd   )-gridCoordinates.x(j,i);
			PetscScalar ry = (Ly-Yd)-gridCoordinates.y(j,i);
			PetscScalar r = PetscSqrtScalar(rx*rx+ry*ry);
			phi_ud = 0.5 + 0.5 * PetscTanhScalar( (Rd-r) / xi );
		}

		//Add
		phiWall(j,i) = PetscMax(phi_b,phi_ud+phi_ld);
	}
	PetscWrapper::RestorePetscArray(da,gridCoordinates);
	Field2D::RestoreFieldArray(phiWall);

	//Write
	WriteFieldInBinaryFormat(phiWall);
	WriteFieldInVTSFormat(phiWall);

	//Destroy
	Field2D::DestroyField(phiWall);
	DestroyDMDA(da);
	PetscFinalize();
	return(0);
}
