#include "Discretisation2D.h"


//	2D matrix index notation:
//
// 	[	22 12 20 12 22
// 		12 11 10 11 12
// 		20 10 00 10 20
// 		12 11 10 11 12
// 		22 12 02 21 22	]


namespace
{
	const PetscInt X_ = 0; /**< \brief Space direction = 0 */
	const PetscInt Y_ = 1; /**< \brief Space direction = 1 */
	const PetscInt Z_ = 2; /**< \brief Space direction = 2 */

	const PetscInt C_ = 0; /**< \brief Centre index = 0 */
	const PetscInt N_ = 1; /**< \brief North index = 1 */
	const PetscInt S_ = 2; /**< \brief South index = 2 */
	const PetscInt E_ = 3; /**< \brief East index = 3 */
	const PetscInt W_ = 4; /**< \brief West index = 4 */
	const PetscInt F_ = 5; /**< \brief Front index = 5 */
	const PetscInt B_ = 6; /**< \brief Back index = 6 */

	const PetscScalar FOURTH_ORDER_GRAD_27_BY_24_ = 27./24.; /**< \brief Fourth-order gradient coefficient = 27/24 */
	const PetscScalar FOURTH_ORDER_GRAD_01_BY_24_ =  1./24.; /**< \brief Fourth-order gradient coefficient = 1/24 */

	const PetscScalar SECOND_ORDER_LAP_00_ = -4.; /**< \brief Second-order Laplacian coefficient = -4 */
	const PetscScalar SECOND_ORDER_LAP_10_ =  1.; /**< \brief Second-order Laplacian coefficient = 1 */
	const PetscScalar FOURTH_ORDER_LAP_00_ = -20./6.; /**< \brief Fourth-order Laplacian = -20/6 */
	const PetscScalar FOURTH_ORDER_LAP_10_ =   4./6.; /**< \brief Fourth-order Laplacian = 4/6 */
	const PetscScalar FOURTH_ORDER_LAP_11_ =   1./6.; /**< \brief Fourth-order Laplacian = 1/6 */

	const PetscScalar SECOND_ORDER_LAP_000_ =  -6.; /**< \brief Second-order Laplacian coefficient = -6 */
	const PetscScalar SECOND_ORDER_LAP_100_ =   1.; /**< \brief Second-order Laplacian coefficient = 1 */
	const PetscScalar FOURTH_ORDER_LAP_000_ = -24./6.; /**< \brief Fourth-order Laplacian coefficient) = -24/6 */
	const PetscScalar FOURTH_ORDER_LAP_100_ =   2./6.; /**< \brief Fourth-order Laplacian coefficient = 2/6 */
	const PetscScalar FOURTH_ORDER_LAP_110_ =   1./6.; /**< \brief Fourth-order Laplacian coefficient = 1/6 */
	const PetscScalar FIFTH_ORDER_LAP_000_  =-128./30.; /**< \brief Fifth-order Laplacian coefficient = -128/30 */
	const PetscScalar FIFTH_ORDER_LAP_100_  =  14./30.; /**< \brief Fifth-order Laplacian coefficient = 14/30 */
	const PetscScalar FIFTH_ORDER_LAP_110_  =  3. /30.; /**< \brief Fifth-order Laplacian coefficient) = 3/30 */
	const PetscScalar FIFTH_ORDER_LAP_111_  =  1. /30.; /**< \brief Fifth-order Laplacian coefficient = 1/30 */

	const PetscScalar WENO_13_BY_12_ = 13./12.; /**< \brief Weno coefficient = 13/12 */
	const PetscScalar WENO_01_BY_04_ = 1./4.; /**< \brief Weno coefficient = 1/4 */
	const PetscScalar WENO_01_BY_06_ = 1./6.; /**< \brief Weno coefficient = 1/6 */
	const PetscScalar WENO_02_BY_06_ = 2./6.; /**< \brief Weno coefficient = 2/6 */
	const PetscScalar WENO_05_BY_06_ = 5./6.; /**< \brief Weno coefficient = 5/6 */
	const PetscScalar WENO_07_BY_06_ = 7./6.; /**< \brief Weno coefficient = 7/6 */
	const PetscScalar WENO_11_BY_06_ = 11./6.; /**< \brief Weno coefficient = 11/6 */
	const PetscScalar SMALL_ = 1e-8; /**< \brief Arbitrarily small positive value = 1e-8 */


	/** \brief Left array of first-order qx derivatives
	 *  \param[in]  f  two-dimensional array
	 *  \param[in]  j  node index
	 *  \param[in]  i  node index
	 *  \param[out] qx derivatives \f$ q^{-}_{i=0,..,4} = \partial_x f_i \f$ */
	inline PetscErrorCode WenoLeftQx
	(
		const PetscScalar *const *const f,
		PetscInt j,
		PetscInt i,
		PetscScalar qx[5]
	)
	{
		qx[0] = f[j  ][i-2] - f[j  ][i-3];
		qx[1] = f[j  ][i-1] - f[j  ][i-2];
		qx[2] = f[j  ][i  ] - f[j  ][i-1];
		qx[3] = f[j  ][i+1] - f[j  ][i  ];
		qx[4] = f[j  ][i+2] - f[j  ][i+1];
		return(0);
	}


	/** \brief Left array of first-order qy derivatives
	*  \param[in]  f  two-dimensional array
	*  \param[in]  j  node index
	*  \param[in]  i  node index
	*  \param[out] qy derivatives \f$ q^{-}_{i=0,..,4} = \partial_y f_i \f$ */
	inline PetscErrorCode WenoLeftQy
	(
		const PetscScalar *const *const f,
		PetscInt j,
		PetscInt i,
		PetscScalar qy[5]
	)
	{
		qy[0] = f[j-2][i  ] - f[j-3][i  ];
		qy[1] = f[j-1][i  ] - f[j-2][i  ];
		qy[2] = f[j  ][i  ] - f[j-1][i  ];
		qy[3] = f[j+1][i  ] - f[j  ][i  ];
		qy[4] = f[j+2][i  ] - f[j+1][i  ];
		return(0);
	}


	/** \brief Right array of first-order qx derivatives
	 *  \param[in]  f  two-dimensional array
	 *  \param[in]  j  node index
	 *  \param[in]  i  node index
	 *  \param[out] qx derivatives \f$ q^{+}_{i=0,..,4} = \partial_x f_i \f$ */
	inline PetscErrorCode WenoRightQx
	(
		const PetscScalar *const *const f,
		PetscInt j,
		PetscInt i,
		PetscScalar qx[5]
	)
	{
		qx[0] = f[j  ][i+3] - f[j  ][i+2];
		qx[1] = f[j  ][i+2] - f[j  ][i+1];
		qx[2] = f[j  ][i+1] - f[j  ][i  ];
		qx[3] = f[j  ][i  ] - f[j  ][i-1];
		qx[4] = f[j  ][i-1] - f[j  ][i-2];
		return(0);
	}


	/** \brief Right array of first-order qy derivatives
	 *  \param[in]  f  two-dimensional array
	 *  \param[in]  j  node index
	 *  \param[in]  i  node index
	 *  \param[out] qy derivatives \f$ q^{+}_{i=0,..,4} = \partial_y f_i \f$ */
	inline PetscErrorCode WenoRightQy
	(
		const PetscScalar *const *const f,
		PetscInt j,
		PetscInt i,
		PetscScalar qy[5]
	)
	{
		qy[0] = f[j+3][i  ] - f[j+2][i  ];
		qy[1] = f[j+2][i  ] - f[j+1][i  ];
		qy[2] = f[j+1][i  ] - f[j  ][i  ];
		qy[3] = f[j  ][i  ] - f[j-1][i  ];
		qy[4] = f[j-1][i  ] - f[j-2][i  ];
		return(0);
	}


	/** \brief Weighted Weno derivative
	 *  \param[in]  q array of first-order derivatives
	 *  \param[out] f derivative \f$ (\partial_\alpha f)_{\alpha=x,y} = \sum w_i \left(\sum \alpha_j q_j \right) \f$ */
	PetscErrorCode WenoWeightedDerivative
	(
		PetscScalar q[5],
		PetscScalar &f
	)
	{
		PetscScalar S0 = WENO_13_BY_12_ * PetscSqr( q[0] - 2.*q[1] + q[2] ) + WENO_01_BY_04_ * PetscSqr(    q[0] - 4.*q[1] + 3.*q[2] );
		PetscScalar S1 = WENO_13_BY_12_ * PetscSqr( q[1] - 2.*q[2] + q[3] ) + WENO_01_BY_04_ * PetscSqr(    q[1]           -    q[3] );
		PetscScalar S2 = WENO_13_BY_12_ * PetscSqr( q[2] - 2.*q[3] + q[4] ) + WENO_01_BY_04_ * PetscSqr( 3.*q[2] - 4.*q[3] +    q[4] );
		PetscScalar omega0 = 0.1 / PetscSqr(1.e-6 + S0);
		PetscScalar omega1 = 0.6 / PetscSqr(1.e-6 + S1);
		PetscScalar omega2 = 0.3 / PetscSqr(1.e-6 + S2);
		PetscScalar omegas = omega0 + omega1 + omega2;
		omega0 /= omegas;
		omega1 /= omegas;
		omega2 /= omegas;
		f =
		(
			  omega0 * (  WENO_02_BY_06_ * q[0] - WENO_07_BY_06_ * q[1] + WENO_11_BY_06_ * q[2] )
			+ omega1 * ( -WENO_01_BY_06_ * q[1] + WENO_05_BY_06_ * q[2] + WENO_02_BY_06_ * q[3] )
			+ omega2 * (  WENO_02_BY_06_ * q[2] + WENO_05_BY_06_ * q[3] - WENO_01_BY_06_ * q[4] )
		);
		return(0);
	}


}


PetscScalar Discretisation2D::SecondOrder::XGrad
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		f[j][i] - f[j][i-1]
	);
}


PetscScalar Discretisation2D::SecondOrder::YGrad
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		f[j][i] - f[j-1][i]
	);
}


PetscScalar Discretisation2D::SecondOrder::Grad
(
	const PetscScalar *const *const f,
	PetscInt dim,
	PetscInt j,
	PetscInt i
)
{
	return Discretisation2D::SecondOrder::GradPtr[dim](f,j,i);
}


PetscScalar Discretisation2D::FourthOrder::XGrad
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  FOURTH_ORDER_GRAD_27_BY_24_ * ( f[j][i  ] - f[j][i-1] )
		- FOURTH_ORDER_GRAD_01_BY_24_ * ( f[j][i+1] - f[j][i-2] )
	);
}


PetscScalar Discretisation2D::FourthOrder::YGrad
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  FOURTH_ORDER_GRAD_27_BY_24_ * ( f[j  ][i] - f[j-1][i] )
		- FOURTH_ORDER_GRAD_01_BY_24_ * ( f[j+1][i] - f[j-2][i] )
	);
}


PetscScalar Discretisation2D::FourthOrder::Grad
(
	const PetscScalar *const *const f,
	PetscInt dim,
	PetscInt j,
	PetscInt i
)
{
	return Discretisation2D::FourthOrder::GradPtr[dim](f,j,i);
}


PetscScalar Discretisation2D::SecondOrder::XGrad2
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		( f[j][i+1]
		- f[j][i  ]
		- f[j][i-1]
		+ f[j][i-2] ) * 0.5
	);
}


PetscScalar Discretisation2D::SecondOrder::YGrad2
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		( f[j+1][i]
		- f[j  ][i]
		- f[j-1][i]
		+ f[j-2][i] ) * 0.5
	);
}


PetscScalar Discretisation2D::SecondOrder::Grad2
(
	const PetscScalar *const *const f,
	PetscInt dim,
	PetscInt j,
	PetscInt i
)
{
	return Discretisation2D::SecondOrder::Grad2Ptr[dim](f,j,i);
}


PetscScalar Discretisation2D::SecondOrder::Divergent
(
	const PetscScalar *const *const fx,
	const PetscScalar *const *const fy,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  Discretisation2D::SecondOrder::XGrad(fx,j  ,i+1)
		+ Discretisation2D::SecondOrder::YGrad(fy,j+1,i  )
	);
}


PetscScalar Discretisation2D::FourthOrder::Divergent
(
	const PetscScalar *const *const fx,
	const PetscScalar *const *const fy,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  Discretisation2D::FourthOrder::XGrad(fx,j  ,i+1)
		+ Discretisation2D::FourthOrder::YGrad(fy,j+1,i  )
	);
}


PetscScalar Discretisation2D::SecondOrder::Divergent
(
	const PetscScalar *const *const *const F,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  Discretisation2D::SecondOrder::XGrad(F[X_],j  ,i+1)
		+ Discretisation2D::SecondOrder::YGrad(F[Y_],j+1,i  )
	);
}


PetscScalar Discretisation2D::FourthOrder::Divergent
(
	const PetscScalar *const *const *const F,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  Discretisation2D::FourthOrder::XGrad(F[X_],j  ,i+1)
		+ Discretisation2D::FourthOrder::YGrad(F[Y_],j+1,i  )
	);
}


PetscScalar Discretisation2D::SecondOrder::Laplacian
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  SECOND_ORDER_LAP_00_ * ( f[j  ][i  ] )
		+ SECOND_ORDER_LAP_10_ * ( f[j+1][i  ]
		                         + f[j-1][i  ]
		                         + f[j  ][i+1]
		                         + f[j  ][i-1] )
	);
}


PetscScalar Discretisation2D::FourthOrder::Laplacian
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  FOURTH_ORDER_LAP_00_ * ( f[j  ][i  ] )
		+ FOURTH_ORDER_LAP_10_ * ( f[j+1][i  ]
		                         + f[j-1][i  ]
		                         + f[j  ][i+1]
		                         + f[j  ][i-1] )
		+ FOURTH_ORDER_LAP_11_ * ( f[j+1][i+1]
		                         + f[j+1][i-1]
		                         + f[j-1][i+1]
		                         + f[j-1][i-1] )
	);
}


PetscScalar Discretisation2D::SecondOrder::Laplacian
(
	const PetscScalar *const eta,
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  eta[C_] * f[j  ][i  ]
		+ eta[N_] * f[j+1][i  ]
		+ eta[S_] * f[j-1][i  ]
		+ eta[E_] * f[j  ][i+1]
		+ eta[W_] * f[j  ][i-1]
	);
}


PetscScalar Discretisation2D::SecondOrder::XTransposedLaplacian
(
	const PetscScalar *const eta,
	const PetscScalar *const *const *const F,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		+ eta[N_] * ( F[Y_][j+1][i  ] - F[Y_][j+1][i-1] )
		- eta[S_] * ( F[Y_][j  ][i  ] - F[Y_][j  ][i-1] )
		+ eta[E_] * ( F[X_][j  ][i+1] - F[X_][j  ][i  ] )
		- eta[W_] * ( F[X_][j  ][i  ] - F[X_][j  ][i-1] )
	);
}


PetscScalar Discretisation2D::SecondOrder::YTransposedLaplacian
(
	const PetscScalar *const eta,
	const PetscScalar *const *const *const F,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		+ eta[N_] * ( F[Y_][j+1][i  ] - F[Y_][j  ][i  ] )
		- eta[S_] * ( F[Y_][j  ][i  ] - F[Y_][j-1][i  ] )
		+ eta[E_] * ( F[X_][j  ][i+1] - F[X_][j-1][i+1] )
		- eta[W_] * ( F[X_][j  ][i  ] - F[X_][j-1][i  ] )
	);
}


PetscScalar Discretisation2D::SecondOrder::TransposedLaplacian
(
	const PetscScalar *const eta,
	const PetscScalar *const *const *const F,
	PetscInt dim,
	PetscInt j,
	PetscInt i
)
{
	return Discretisation2D::SecondOrder::TransposedLaplacianPtr[dim](eta,F,j,i);
}


PetscErrorCode Discretisation2D::SecondOrder::IMinusLaplacian
(
	const PetscScalar& theta,
	PetscInt j,
	PetscInt i,
	MatStencil &row,
	MatStencil  col[5],
	PetscScalar vec[5]
)
{
	row.j = j;
	row.i = i;
	vec[0] = 1.0 - theta * SECOND_ORDER_LAP_00_; col[0].j = (j  ); col[0].i = (i  );
	vec[1] =     - theta * SECOND_ORDER_LAP_10_; col[1].j = (j+1); col[1].i = (i  );
	vec[2] =     - theta * SECOND_ORDER_LAP_10_; col[2].j = (j-1); col[2].i = (i  );
	vec[3] =     - theta * SECOND_ORDER_LAP_10_; col[3].j = (j  ); col[3].i = (i+1);
	vec[4] =     - theta * SECOND_ORDER_LAP_10_; col[4].j = (j  ); col[4].i = (i-1);
	return(0);
}


PetscErrorCode Discretisation2D::SecondOrder::IMinusLaplacian
(
	const PetscScalar& theta,
	const PetscScalar *const eta,
	PetscInt j,
	PetscInt i,
	MatStencil &row,
	MatStencil  col[5],
	PetscScalar vec[5]
)
{
	row.j = j;
	row.i = i;
	vec[0] = 1.0 - theta * ( eta[C_] ); col[0].j = (j  ); col[0].i = (i  );
	vec[1] =     - theta * ( eta[N_] ); col[1].j = (j+1); col[1].i = (i  );
	vec[2] =     - theta * ( eta[S_] ); col[2].j = (j-1); col[2].i = (i  );
	vec[3] =     - theta * ( eta[E_] ); col[3].j = (j  ); col[3].i = (i+1);
	vec[4] =     - theta * ( eta[W_] ); col[4].j = (j  ); col[4].i = (i-1);
	return(0);
}


PetscErrorCode Discretisation2D::FourthOrder::IMinusLaplacian
(
	const PetscScalar& theta,
	PetscInt j,
	PetscInt i,
	MatStencil &row,
	MatStencil  col[9],
	PetscScalar vec[9]
)
{
	row.j = j;
	row.i = i;
	vec[0] = 1.0 - theta * FOURTH_ORDER_LAP_00_; col[0].j = (j  ); col[0].i = (i  );
	vec[1] =     - theta * FOURTH_ORDER_LAP_10_; col[1].j = (j+1); col[1].i = (i  );
	vec[2] =     - theta * FOURTH_ORDER_LAP_10_; col[2].j = (j-1); col[2].i = (i  );
	vec[3] =     - theta * FOURTH_ORDER_LAP_10_; col[3].j = (j  ); col[3].i = (i+1);
	vec[4] =     - theta * FOURTH_ORDER_LAP_10_; col[4].j = (j  ); col[4].i = (i-1);
	vec[5] =     - theta * FOURTH_ORDER_LAP_11_; col[5].j = (j+1); col[5].i = (i+1);
	vec[6] =     - theta * FOURTH_ORDER_LAP_11_; col[6].j = (j+1); col[6].i = (i-1);
	vec[7] =     - theta * FOURTH_ORDER_LAP_11_; col[7].j = (j-1); col[7].i = (i+1);
	vec[8] =     - theta * FOURTH_ORDER_LAP_11_; col[8].j = (j-1); col[8].i = (i-1);
	return(0);
}


PetscScalar Discretisation2D::FirstOderUpwind::UGrad
(
	const PetscScalar *const uDir,
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	PetscScalar fx,fy;
	if      (uDir[X_] < -SMALL_) { fx = ( f[j  ][i+1] - f[j  ][i  ] ); }
	else if (uDir[X_] >  SMALL_) { fx = ( f[j  ][i  ] - f[j  ][i-1] ); }
	else    { fx = 0.; }
	if      (uDir[Y_] < -SMALL_) { fy = ( f[j+1][i  ] - f[j  ][i  ] ); }
	else if (uDir[Y_] >  SMALL_) { fy = ( f[j  ][i  ] - f[j-1][i  ] ); }
	else    { fy = 0.; }
	return
	(
		  uDir[X_] * fx
		+ uDir[Y_] * fy
	);
}


PetscScalar Discretisation2D::SecondOderUpwind::UGrad
(
	const PetscScalar *const uDir,
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	PetscScalar fx,fy;
	if      (uDir[X_] < -SMALL_) { fx = (-0.5 * f[j  ][i+2] + 2. * f[j  ][i+1] - 1.5 * f[j  ][i  ] ); }
	else if (uDir[X_] >  SMALL_) { fx = ( 1.5 * f[j  ][i  ] - 2. * f[j  ][i-1] - 0.5 * f[j  ][i-2] ); }
	else    { fx = 0.; }
	if      (uDir[Y_] < -SMALL_) { fy = (-0.5 * f[j+2][i  ] + 2. * f[j+1][i  ] - 1.5 * f[j  ][i  ] ); }
	else if (uDir[Y_] >  SMALL_) { fy = ( 1.5 * f[j  ][i  ] - 2. * f[j-1][i  ] - 0.5 * f[j-2][i  ] ); }
	else    { fy = 0.; }
	return
	(
		  uDir[X_] * fx
		+ uDir[Y_] * fy
	);
}


PetscScalar Discretisation2D::Weno::UGrad
(
	const PetscScalar *const  uDir,
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	PetscScalar q[5];
	PetscScalar fx,fy;
	if      (uDir[X_] < -SMALL_){ WenoRightQx(f,j,i,q); WenoWeightedDerivative(q,fx); }
	else if (uDir[X_] >  SMALL_){ WenoLeftQx (f,j,i,q); WenoWeightedDerivative(q,fx); }
	else    { fx = 0.; }
	if      (uDir[Y_] < -SMALL_){ WenoRightQy(f,j,i,q); WenoWeightedDerivative(q,fy); }
	else if (uDir[Y_] >  SMALL_){ WenoLeftQy (f,j,i,q); WenoWeightedDerivative(q,fy); }
	else    { fy = 0.; }
	return
	(
		  uDir[X_] * fx
		+ uDir[Y_] * fy
	);
}


// 	const PetscScalar SECOND_ORDER_BIH_00_ = 20.; /**< \brief Second-order biharmonic coefficient = 20 */
// 	const PetscScalar SECOND_ORDER_BIH_10_ = -8.; /**< \brief Second-order biharmonic coefficient = -8 */
// 	const PetscScalar SECOND_ORDER_BIH_11_ =  2.; /**< \brief Second-order biharmonic coefficient =  2 */
// 	const PetscScalar SECOND_ORDER_BIH_20_ =  1.; /**< \brief Second-order biharmonic coefficient =  1 */
// 	const PetscScalar FOURTH_ORDER_BIH_00_ =  468./36.; /**< \brief Fourth-order biharmonic coefficient =  468/36 */
// 	const PetscScalar FOURTH_ORDER_BIH_10_ = -144./36.; /**< \brief Fourth-order biharmonic coefficient = -144/36 */
// 	const PetscScalar FOURTH_ORDER_BIH_11_ =   -8./36.; /**< \brief Fourth-order biharmonic coefficient =   -8/36 */
// 	const PetscScalar FOURTH_ORDER_BIH_20_ =   18./36.; /**< \brief Fourth-order biharmonic coefficient =   18/36 */
// 	const PetscScalar FOURTH_ORDER_BIH_21_ =    8./36.; /**< \brief Fourth-order biharmonic coefficient =    8/36 */
// 	const PetscScalar FOURTH_ORDER_BIH_22_ =    1./36.; /**< \brief Fourth-order biharmonic coefficient =    1/36 */


// PetscScalar Discretisation2D::SecondOrder::Biharmonic
// (
// 	const PetscScalar *const *const f,
// 	PetscInt j,
// 	PetscInt i
// )
// {
// 	return
// 	(
// 		  SECOND_ORDER_BIH_00_ * ( f[j  ][i  ] )
//
// 		+ SECOND_ORDER_BIH_10_ * ( f[j+1][i  ]
// 		                         + f[j-1][i  ]
// 		                         + f[j  ][i+1]
// 		                         + f[j  ][i-1] )
//
// 		+ SECOND_ORDER_BIH_11_ * ( f[j+1][i+1]
// 		                         + f[j+1][i-1]
// 		                         + f[j-1][i+1]
// 		                         + f[j-1][i-1] )
//
// 		+ SECOND_ORDER_BIH_20_ * ( f[j+2][i  ]
// 		                         + f[j-2][i  ]
// 		                         + f[j  ][i+2]
// 		                         + f[j  ][i-2] )
// 	);
// }


// PetscErrorCode Discretisation2D::SecondOrder::IMinusBiharmonic
// (
// 	const PetscScalar& theta,
// 	PetscInt j,
// 	PetscInt i,
// 	MatStencil &row,
// 	MatStencil  col[13],
// 	PetscScalar vec[13]
// )
// {
// 	row.j = j;
// 	row.i = i;
//
// 	vec[0 ] = 1.0 - theta * SECOND_ORDER_BIH_00_; col[0 ].j = (j  ); col[0 ].i = (i  );
//
// 	vec[1 ] =     - theta * SECOND_ORDER_BIH_10_; col[1 ].j = (j+1); col[1 ].i = (i  );
// 	vec[2 ] =     - theta * SECOND_ORDER_BIH_10_; col[2 ].j = (j-1); col[2 ].i = (i  );
// 	vec[3 ] =     - theta * SECOND_ORDER_BIH_10_; col[3 ].j = (j  ); col[3 ].i = (i+1);
// 	vec[4 ] =     - theta * SECOND_ORDER_BIH_10_; col[4 ].j = (j  ); col[4 ].i = (i-1);
//
// 	vec[5 ] =     - theta * SECOND_ORDER_BIH_11_; col[5 ].j = (j+1); col[5 ].i = (i+1);
// 	vec[6 ] =     - theta * SECOND_ORDER_BIH_11_; col[6 ].j = (j+1); col[6 ].i = (i-1);
// 	vec[7 ] =     - theta * SECOND_ORDER_BIH_11_; col[7 ].j = (j-1); col[7 ].i = (i+1);
// 	vec[8 ] =     - theta * SECOND_ORDER_BIH_11_; col[8 ].j = (j-1); col[8 ].i = (i-1);
//
// 	vec[9 ] =     - theta * SECOND_ORDER_BIH_20_; col[9 ].j = (j+2); col[9 ].i = (i  );
// 	vec[10] =     - theta * SECOND_ORDER_BIH_20_; col[10].j = (j-2); col[10].i = (i  );
// 	vec[11] =     - theta * SECOND_ORDER_BIH_20_; col[11].j = (j  ); col[11].i = (i+2);
// 	vec[12] =     - theta * SECOND_ORDER_BIH_20_; col[12].j = (j  ); col[12].i = (i-2);
// 	return(0);
// }


// PetscScalar Discretisation2D::FourthOrder::Biharmonic
// (
// 	const PetscScalar *const *const f,
// 	PetscInt j,
// 	PetscInt i
// )
// {
// 	return
// 	(
// 		  FOURTH_ORDER_BIH_00_ * ( f[j  ][i  ] )
//
// 		+ FOURTH_ORDER_BIH_10_ * ( f[j+1][i  ]
// 		                         + f[j-1][i  ]
// 		                         + f[j  ][i+1]
// 		                         + f[j  ][i-1] )
//
// 		+ FOURTH_ORDER_BIH_11_ * ( f[j+1][i+1]
// 		                         + f[j+1][i-1]
// 		                         + f[j-1][i+1]
// 		                         + f[j-1][i-1] )
//
// 		+ FOURTH_ORDER_BIH_20_ * ( f[j+2][i  ]
// 		                         + f[j-2][i  ]
// 		                         + f[j  ][i+2]
// 		                         + f[j  ][i-2] )
//
// 		+ FOURTH_ORDER_BIH_21_ * ( f[j+2][i+1]
// 		                         + f[j+2][i-1]
// 		                         + f[j-2][i+1]
// 		                         + f[j-2][i-1]
// 		                         + f[j+1][i+2]
// 		                         + f[j+1][i-2]
// 		                         + f[j-1][i+2]
// 		                         + f[j-1][i-2] )
//
// 		+ FOURTH_ORDER_BIH_22_ * ( f[j+2][i+2]
// 		                         + f[j+2][i-2]
// 		                         + f[j-2][i+2]
// 		                         + f[j-2][i-2] )
// 	);
// }


// PetscErrorCode Discretisation2D::FourthOrder::IMinusBiharmonic
// (
// 	const PetscScalar& theta,
// 	PetscInt j,
// 	PetscInt i,
// 	MatStencil &row,
// 	MatStencil  col[25],
// 	PetscScalar vec[25]
// )
// {
// 	row.j = j;
// 	row.i = i;
//
// 	vec[0 ] = 1. - theta * FOURTH_ORDER_BIH_00_; col[0 ].j = (j  ); col[0 ].i = (i  );
//
// 	vec[1 ] =    - theta * FOURTH_ORDER_BIH_10_; col[1 ].j = (j+1); col[1 ].i = (i  );
// 	vec[2 ] =    - theta * FOURTH_ORDER_BIH_10_; col[2 ].j = (j-1); col[2 ].i = (i  );
// 	vec[3 ] =    - theta * FOURTH_ORDER_BIH_10_; col[3 ].j = (j  ); col[3 ].i = (i+1);
// 	vec[4 ] =    - theta * FOURTH_ORDER_BIH_10_; col[4 ].j = (j  ); col[4 ].i = (i-1);
//
// 	vec[5 ] =    - theta * FOURTH_ORDER_BIH_11_; col[5 ].j = (j+1); col[5 ].i = (i+1);
// 	vec[6 ] =    - theta * FOURTH_ORDER_BIH_11_; col[6 ].j = (j+1); col[6 ].i = (i-1);
// 	vec[7 ] =    - theta * FOURTH_ORDER_BIH_11_; col[7 ].j = (j-1); col[7 ].i = (i+1);
// 	vec[8 ] =    - theta * FOURTH_ORDER_BIH_11_; col[8 ].j = (j-1); col[8 ].i = (i-1);
//
// 	vec[9 ] =    - theta * FOURTH_ORDER_BIH_20_; col[9 ].j = (j+2); col[9 ].i = (i  );
// 	vec[10] =    - theta * FOURTH_ORDER_BIH_20_; col[10].j = (j-2); col[10].i = (i  );
// 	vec[11] =    - theta * FOURTH_ORDER_BIH_20_; col[11].j = (j  ); col[11].i = (i+2);
// 	vec[12] =    - theta * FOURTH_ORDER_BIH_20_; col[12].j = (j  ); col[12].i = (i-2);
//
// 	vec[13] =    - theta * FOURTH_ORDER_BIH_21_; col[13].j = (j+2); col[13].i = (i+1);
// 	vec[14] =    - theta * FOURTH_ORDER_BIH_21_; col[14].j = (j+2); col[14].i = (i-1);
// 	vec[15] =    - theta * FOURTH_ORDER_BIH_21_; col[15].j = (j-2); col[15].i = (i+1);
// 	vec[16] =    - theta * FOURTH_ORDER_BIH_21_; col[16].j = (j-2); col[16].i = (i-1);
// 	vec[17] =    - theta * FOURTH_ORDER_BIH_21_; col[17].j = (j+1); col[17].i = (i+2);
// 	vec[18] =    - theta * FOURTH_ORDER_BIH_21_; col[18].j = (j+1); col[18].i = (i-2);
// 	vec[19] =    - theta * FOURTH_ORDER_BIH_21_; col[19].j = (j-1); col[19].i = (i+2);
// 	vec[20] =    - theta * FOURTH_ORDER_BIH_21_; col[20].j = (j-1); col[20].i = (i-2);
//
// 	vec[21] =    - theta * FOURTH_ORDER_BIH_22_; col[21].j = (j+2); col[21].i = (i+2);
// 	vec[22] =    - theta * FOURTH_ORDER_BIH_22_; col[22].j = (j+2); col[22].i = (i-2);
// 	vec[23] =    - theta * FOURTH_ORDER_BIH_22_; col[23].j = (j-2); col[23].i = (i+2);
// 	vec[24] =    - theta * FOURTH_ORDER_BIH_22_; col[24].j = (j-2); col[24].i = (i-2);
// 	return(0);
// }
