#ifndef DISCRETISATION2D_H
#define DISCRETISATION2D_H


/** \file
 *  \brief Collection of space discretisation schemes */


#include "petsc.h"


namespace
{


	typedef PetscScalar(*TwoDimensionalGradOperator)(const PetscScalar *const *const,PetscInt,PetscInt);
	typedef PetscScalar(*TwoDimensionalGrad2Operator)(const PetscScalar *const *const,PetscInt,PetscInt);
	typedef PetscScalar(*TwoDimensionalTransposedLaplacianOperator)(const PetscScalar *const, const PetscScalar *const *const *const,PetscInt,PetscInt);


}


/** \namespace Discretisation2D
 *  \brief Collection of space discretisation schemes */
namespace Discretisation2D
{


	/** \namespace Discretisation2D::SecondOrder
	 *  \brief Collection of second order space discretisation schemes */
	namespace SecondOrder
	{


		/** \brief 2-point stencil Grad_x(f)
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_x f \right|_{j,i-1/2} \f$ */
		PetscScalar XGrad
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil Grad_y(f)
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_y f \right|_{j-1/2,i} \f$ */
		PetscScalar YGrad
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to second order %Grad */
		const TwoDimensionalGradOperator GradPtr[] =
		{
			Discretisation2D::SecondOrder::XGrad,
			Discretisation2D::SecondOrder::YGrad
		};


		/** \brief 2-point stencil Grad(f)
		 *  \param[in] f   two-dimensional array
		 *  \param[in] dim space index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        scalar \f$ \left. \partial_x f \right|_{j    ,i-1/2} \f$ if dim = 0 else
		*                         \f$ \left. \partial_y f \right|_{j-1/2,    i} \f$ if dim = 1 */
		PetscScalar Grad
		(
			const PetscScalar *const *const f,
			PetscInt dim,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil Grad_xx(f)
		 *  \note http://web.media.mit.edu/~crtaylor/calculator.html
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_{xx} f \right|_{j,i-1/2} \f$ */
		PetscScalar XGrad2
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil Grad_yy(f)
		 *  \note http://web.media.mit.edu/~crtaylor/calculator.html
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_{yy} f \right|_{j-1/2,i} \f$ */
		PetscScalar YGrad2
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to second order %Grad_ii */
		const TwoDimensionalGrad2Operator Grad2Ptr[] =
		{
			Discretisation2D::SecondOrder::XGrad2,
			Discretisation2D::SecondOrder::YGrad2
		};


		/** \brief 4-point stencil Grad_ii(f)
		 *  \note http://web.media.mit.edu/~crtaylor/calculator.html
		 *  \param[in] f   two-dimensional array
		 *  \param[in] dim space dimension
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        scalar \f$ \left. \partial_{xx} f \right|_{j    ,i-1/2} \f$ if dim = 0 else
		 *                        \f$ \left. \partial_{yy} f \right|_{j-1/2,i    } \f$ if dim = 1 */
		PetscScalar Grad2
		(
			const PetscScalar *const *const f,
			PetscInt dim,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil divergent Div(fx,fy)
		 *  \note \f$ \textbf{f} = ( f_x|_{j,i-1/2} , f_y|_{j-1/2,i} )^\top \f$
		 *  \note \f$ \partial_x f_x |_{j,i} = f_x|_{j    ,i+1/2} - f_x|_{j    ,i-1/2} = f_x|_{j  ,I+1} - f_x|_{j,I} \f$
		 *  \note \f$ \partial_y f_y |_{j,i} = f_y|_{j+1/2,i    } - f_y|_{j-1/2,i    } = f_y|_{J+1,i  } - f_y|_{J,i} \f$
		 *  \param[in] fx two-dimensional arrays \f$ f_x|_{j    ,i-1/2} \f$
		 *  \param[in] fy two-dimensional arrays \f$ f_y|_{j-1/2,i    } \f$
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \nabla \cdot \textbf{f} \right|_{j,i} = \partial_x f_x + \partial_y f_y \f$ */
		PetscScalar Divergent
		(
			const PetscScalar *const *const fx,
			const PetscScalar *const *const fy,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil divergent Div(F)
		 *  \note \f$ \textbf{f} = ( f_x|_{j,i-1/2} , f_y|_{j-1/2,i} )^\top \f$
		 *  \note \f$ \partial_x f_x |_{j,i} = f_x|_{j    ,i+1/2} - f_x|_{j    ,i-1/2} = f_x|_{j  ,I+1} - f_x|_{j,I} \f$
		 *  \note \f$ \partial_y f_y |_{j,i} = f_y|_{j+1/2,i    } - f_y|_{j-1/2,i    } = f_y|_{J+1,i  } - f_y|_{J,i} \f$
		 *  \param[in] F array of two-dimensional arrays \f$ \textbf{f} = ( f_x|_{j,i-1/2},f_y|_{j-1/2,i} )^\top \f$
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \nabla \cdot \textbf{f} \right|_{j,i} = \partial_x f_x + \partial_y f_y \f$ */
		PetscScalar Divergent
		(
			const PetscScalar *const *const *const F,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 5-point stencil Laplacian Lap(f)
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \nabla^2 f \right|_{j,i} \f$ */
		PetscScalar Laplacian
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 5-point stencil Laplacian Lap(eta,f) = Div( eta Grad(f) )
		 *  \param[in] eta interpolated viscosity list \f$ \left[ \eta_C, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] f   two-dimensional array
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        scalar \f$ \left. \nabla \cdot \left( \eta \nabla f \right) \right|_{j,i} \f$ */
		PetscScalar Laplacian
		(
			const PetscScalar *const eta,
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 5-point stencil x-component of transposed Lalacian Div( eta [Grad(F)]^T )
		 *  \param[in] eta one dimensional array of interpolated viscosities \f$ \left[ \eta_C, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] F   array of two-dimensional arrays \f$ \textbf{f} = ( f_x|_{j,i-1/2},f_y|_{j-1/2,i} )^\top \f$
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        x-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{j,i-1/2} \f$ */
		PetscScalar XTransposedLaplacian
		(
			const PetscScalar *const eta,
			const PetscScalar *const *const *const F,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 5-point stencil y-component of transposed Laplacian Div( eta [Grad(F)]^T )
		 *  \param[in] eta one dimensional array of interpolated viscosities \f$ \left[ \eta_C, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] F   array of two-dimensional arrays \f$ \textbf{f} = ( f_x|_{j,i-1/2},f_y|_{j-1/2,i} )^\top \f$
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        y-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{j-1/2,i} \f$ */
		PetscScalar YTransposedLaplacian
		(
			const PetscScalar *const  eta,
			const PetscScalar *const *const *const F,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to transposed Laplacian */
		const TwoDimensionalTransposedLaplacianOperator TransposedLaplacianPtr[] =
		{
			Discretisation2D::SecondOrder::XTransposedLaplacian,
			Discretisation2D::SecondOrder::YTransposedLaplacian
		};


		/** \brief Compact 5-point stencil of transposed Laplacian Div( eta [Grad(F)]^T )
		 *  \param[in] eta one dimensional array of interpolated viscosities \f$ \left[ \eta_C, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] F   array of two-dimensional arrays \f$ \textbf{f} = ( f_x|_{j,i-1/2},f_y|_{j-1/2,i} )^\top \f$
		 *  \param[in] dim space dimension
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  return         x-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{j    ,i-1/2} \f$ if dim = 0 else
		 *                 y-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{j-1/2,i    } \f$ if dim = 1 */
		PetscScalar TransposedLaplacian
		(
			const PetscScalar *const  eta,
			const PetscScalar *const *const *const F,
			PetscInt dim,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 5-point stencil implicit Laplacian LHS = [ I - theta . Lap ]
		 *  \param[in]  theta scalar
		 *  \param[in]  j     node index
		 *  \param[in]  i     node index
		 *  \param[out] row   Petsc matrix stencil
		 *  \param[out] col   Petsc matrix stencil
		 *  \param[out] vec   one dimensional array \f$ \delta_{ji} - \theta \left. \nabla^2 \right|_{j,i} \f$ */
		PetscErrorCode IMinusLaplacian
		(
			const PetscScalar& theta,
			PetscInt j,
			PetscInt i,
			MatStencil &row,
			MatStencil  col[5],
			PetscScalar vec[5]
		);


		/** \brief Compact 5-point stencil implicit Laplacian LHS = [ I - theta . Div( eta Grad ) ]
		 *  \param[in]  theta scalar
		 *  \param[in]  eta   interpolated viscosity array \f$ \left[ \eta_C, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in]  j     node index
		 *  \param[in]  i     node index
		 *  \param[out] row   Petsc matrix stencil
		 *  \param[out] col   Petsc matrix stencil
		 *  \param[out] vec   one dimensional array \f$ \delta_{ji} - \theta \left. \nabla^2 \eta \right|_{j,i} \f$ */
		PetscErrorCode IMinusLaplacian
		(
			const PetscScalar& theta,
			const PetscScalar *const eta,
			PetscInt j,
			PetscInt i,
			MatStencil &row,
			MatStencil  col[5],
			PetscScalar vec[5]
		);


	}


	/** \namespace Discretisation2D::FourthOrder
	 *  \brief Collection of fourth order central space discretisation schemes */
	namespace FourthOrder
	{


		/** \brief 4-point stencil Grad_x(f)
		 *  \note Yefet and Petropoulos, A staggered fourth order accurate explicit finite difference scheme for the time domain Maxwell's equations, JCP, 2001
		 *  \copydetails Discretisation2D::SecondOrder::XGrad */
		PetscScalar XGrad
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil Grad_y(f)
		 *  \note Yefet and Petropoulos, A staggered fourth order accurate explicit finite difference scheme for the time domain Maxwell's equations, JCP, 2001
		 *  \copydetails Discretisation2D::SecondOrder::YGrad */
		PetscScalar YGrad
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to fourth-order gradient
		 *  \copydetails Discretisation2D::SecondOrder::GradPtr */
		const TwoDimensionalGradOperator GradPtr[] =
		{
			Discretisation2D::FourthOrder::XGrad,
			Discretisation2D::FourthOrder::YGrad
		};


		/** \brief 4-point stencil Grad_i(f)
		*   \copydetails Discretisation2D::SecondOrder::Grad */
		PetscScalar Grad
		(
			const PetscScalar *const *const f,
			PetscInt dim,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil divergent Div(F)
		 *  \copydetails Discretisation2D::SecondOrder::Divergent(const PetscScalar *const *const *const,PetscInt,PetscInt) */
		PetscScalar Divergent
		(
			const PetscScalar *const *const *const F,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil divergent Div(fx,fy)
		 *  \copydetails Discretisation2D::SecondOrder::Divergent(const PetscScalar *const *const,const PetscScalar *const *const,PetscInt,PetscInt) */
		PetscScalar Divergent
		(
			const PetscScalar *const *const fx,
			const PetscScalar *const *const fy,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 9-point stencil Laplacian Lap(f)
		 *  \copydetails Discretisation2D::SecondOrder::Laplacian(const PetscScalar *const *const,PetscInt,PetscInt) */
		PetscScalar Laplacian
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 9-point stencil implicit Laplacian LHS = [ I - theta . Lap ]
		*  \copydetails Discretisation2D::SecondOrder::IMinusLaplacian */
		PetscErrorCode IMinusLaplacian
		(
			const PetscScalar& theta,
			PetscInt j,
			PetscInt i,
			MatStencil &row,
			MatStencil  col[9],
			PetscScalar vec[9]
		);


	}


	/** \namespace Discretisation2D::FifthOrder
	 *  \brief Collection of fifth order central space discretisation schemes */
	namespace FifthOrder
	{


		/** \brief Compact 27-point stencil Laplacian Lap(f)
		 *  \note  Spotz, 1996, High-order compact formulation for the 3D poisson equation
		 *  \copydetails Discretisation2D::SecondOrder::Laplacian(const PetscScalar *const *const,PetscInt,PetscInt) */
		PetscScalar Laplacian
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


	}

	/** \namespace Discretisation2D::FirstOderUpwind
	 *  \brief First-order upwind scheme */
	namespace FirstOderUpwind
	{


		/** \brief First-order upwind scheme U.Grad(f)
		 *  \param[in] uDir convective velocity \f$ \textbf{u}|_{j,i} = \left( u_x, u_y \right )^\top \f$
		 *  \param[in] f    two-dimensional array
		 *  \param[in] j    node index
		 *  \param[in] i    node index
		 *  \return         U.Grad(f) = \f$ \left. \textbf{u} \cdot ( \nabla f ) \right|_{j,i} \f$ */
		PetscScalar UGrad
		(
			const PetscScalar *const  uDir,
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);



	}


	/** \namespace Discretisation2D::SecondOderUpwind
	 *  \brief Second-order upwind scheme */
	namespace SecondOderUpwind
	{


		/** \brief Second-order upwind scheme
		 *  \copydetails Discretisation2D::FirstOderUpwind::UGrad */
		PetscScalar UGrad
		(
			const PetscScalar *const  uDir,
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


	}


	/** \namespace Discretisation2D::Weno
	 *  \brief %Weno convection scheme */
	namespace Weno
	{


		/** \brief %Weno scheme
		 *  \copydetails Discretisation2D::FirstOderUpwind::UGrad */
		PetscScalar UGrad
		(
			const PetscScalar *const  uDir,
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


	}


}


#endif


// 		/** \brief 13-point stencil biharmonic Bih(f)
// 		 *  \param[in] f two-dimensional array
// 		 *  \param[in] j node index
// 		 *  \param[in] i node index
// 		 *  \return scalar \f$ \left. \nabla^4 f \right|_{j,i} \f$ */
// 		PetscScalar Biharmonic
// 		(
// 			const PetscScalar *const *const f,
// 			PetscInt j,
// 			PetscInt i
// 		);


// 		/** \brief 13-point stencil implicit biharmonic LHS =  [ I - theta . Bih ]
// 		 *  \param[in]  theta scalar
// 		 *  \param[in]  j     node index
// 		 *  \param[in]  i     node index
// 		 *  \param[out] row   Petsc matrix stencil
// 		 *  \param[out] col   Petsc matrix stencil
// 		 *  \param[out] vec   one dimensional array \f$ \delta_{ji} - \theta \left. \nabla^4 \right|_{j,i} \f$ */
// 		PetscErrorCode IMinusBiharmonic
// 		(
// 			const PetscScalar& theta,
// 			PetscInt j,
// 			PetscInt i,
// 			MatStencil &row,
// 			MatStencil  col[13],
// 			PetscScalar vec[13]
// 		);



// 		/** \brief 25-point stencil biharmonic Bih(f)
// 		 *  \copydetails Discretisation2D::SecondOrder::Biharmonic */
// 		PetscScalar Biharmonic
// 		(
// 			const PetscScalar *const *const f,
// 			PetscInt j,
// 			PetscInt i
// 		);


// 		/** \brief 25-point stencil implicit biharmonic LHS = [ I - theta . Bih ]
// 		 *  \copydetails Discretisation2D::SecondOrder::IMinusBiharmonic */
// 		PetscErrorCode IMinusBiharmonic
// 		(
// 			const PetscScalar& theta,
// 			PetscInt j,
// 			PetscInt i,
// 			MatStencil &row,
// 			MatStencil  col[25],
// 			PetscScalar vec[25]
// 		);
