#include "Constant.h"
#include "PetscWrapper.h"
#include "Field2D.h"
#include "Utils.h"


static char help[] = "Export each single bubble as VTS.\n\n";


PetscErrorCode GetTimes
(
	PetscInt &tmin,
	PetscInt &tmax,
	PetscInt &dt
)
{
	char str[PETSC_MAX_PATH_LEN];
	PetscBool flg;
	PetscOptionsGetString(NULL,NULL,"-t",str,PETSC_MAX_PATH_LEN,&flg);
	if (!flg)
	{
		SETERRQ(PETSC_COMM_WORLD,1,"Pass with -t tmin:tmax:dt");
	}
	char *p = strtok(str,":"); tmin = atoi(p);
	p = strtok(NULL,":"); tmax = (p == NULL) ? tmin+1 : atoi(p);
	p = strtok(NULL,":"); dt = (p == NULL) ? 1 : atoi(p);
	return(0);
}


PetscInt main( int argc, char* argv[] )
{
	PetscInitialize(&argc,&argv,(char*)0,help);
	Constant::ReadInputFile();
	PetscBool exportWall=PETSC_FALSE;
	PetscOptionsGetBool(NULL,NULL,"-wall",&exportWall,NULL);
	DM da;
	DMDACreate2d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		(Constant::Grid::Nx), //global dimension in x-direction of the array
		(Constant::Grid::Ny), //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		2,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		&da                   //dynamic array
	);
	DMSetUp(da);
	DMDASetUniformCoordinates
	(
		da,
		0.,(Constant::Grid::L[0]),
		0.,(Constant::Grid::L[1]),
		0.,(Constant::Grid::L[2])
	);
	PetscInt n;
	Field2D::VectorField phi;
	Field2D::ScalarField phiSum;
	PetscWrapper::ReadScalarInBinaryFormat(n,"bubbleNumber");
	Field2D::CreateField(da,phi,"phi",n);
	Field2D::CreateField(da,phiSum,"phiSum");
	PetscInt tmin,tmax,dt;
	GetTimes(tmin,tmax,dt);
	for (PetscInt t=tmin; t<tmax; t+=dt)
	{
		Field2D::ReadFieldInBinaryFormat(phi,t);
		Field2D::SumVectorField(phi,phiSum);
		Field2D::WriteFieldInVTSFormat(phi,t);
		Field2D::WriteFieldInVTSFormat(phiSum,t);
	}
	Field2D::DestroyField(phi);
	Field2D::DestroyField(phiSum);
	if (exportWall)
	{
		Field2D::ScalarField phiWall;
		Field2D::CreateField(da,phiWall,"phiWall");
		Field2D::ReadFieldInBinaryFormat(phiWall);
		Field2D::WriteFieldInVTSFormat(phiWall);
		Field2D::DestroyField(phiWall);
	}
	DMDestroy(&da);
	PetscFinalize();
	return(0);
}
