#include "Constant.h"
#include "PetscWrapper.h"
#include "Field2D.h"
#include "Utils.h"
#include "Context.h"
#include "SNESWrapper.h"
#include "PhiEquation.h"
#include "PressureEquation.h"
#include "Discretisation2D.h"


static char help[] = "Export each single bubble as VTS.\n\n";


PetscErrorCode GetTimes
(
	PetscInt &tmin,
	PetscInt &tmax,
	PetscInt &dt
)
{
	char str[PETSC_MAX_PATH_LEN];
	PetscBool flg;
	PetscOptionsGetString(NULL,NULL,"-t",str,PETSC_MAX_PATH_LEN,&flg);
	if (!flg)
	{
		SETERRQ(PETSC_COMM_WORLD,1,"Pass with -t tmin:tmax:dt");
	}
	char *p = strtok(str,":"); tmin = atoi(p);
	p = strtok(NULL,":"); tmax = (p == NULL) ? tmin+1 : atoi(p);
	p = strtok(NULL,":"); dt = (p == NULL) ? 1 : atoi(p);
	return(0);
}


PetscErrorCode ComputeGasPressure
(
	Context::Ctx &ctx,
	SNESWrapper::PetscNonLinearSolver &pressureSolver
)
{
// 	//Print volume
// 	PhiEquation::ComputeBubbleVolume(ctx);
// 	ForAllB(ctx.phi.n)
// 	{
// 		PetscPrintf(PETSC_COMM_WORLD,"  V[%i] = %4.2f \n",b,ctx.bubbleVolume[b]);
// 	}

	//Compute surface force
	PhiEquation::ComputeSurfaceForce(ctx);

	//Loop over each bubble
	Field2D::GetGhostedFieldArray(ctx.sFx);
	Field2D::GetGhostedFieldArray(ctx.sFy);
	ForAllB(ctx.phi.n)
	{
		//Compute Div(surface force)
		PetscScalar alpha = -(3. * PetscSqrtScalar(2.) ) * (ctx.xi) * (ctx.invDx3);
		Field2D::GetFieldArray(ctx.divRHS);
		ForAllJI(ctx.da)
		{
			ctx.divRHS(j,i) =
			(
				Discretisation2D::SecondOrder::Divergent(ctx.sFx(b),ctx.sFy(b),j,i) * alpha
			); //(1/dx) moved to pressure equation
		}
		Field2D::RestoreFieldArray(ctx.divRHS);

		//Compute gas pressure
		SNESWrapper::SetNonLinearSolverSolution(pressureSolver,ctx.pGas.g[b]);
		SNESWrapper::RunPetscNonLinearSolver
		(
			pressureSolver,
			PressureEquation::FormSNESFunction,
			PressureEquation::FormSNESJacobian,
			&ctx
		);
		SNESWrapper::GetNonLinearSolverSolution(pressureSolver,ctx.pGas.g[b]);
	}
	Field2D::RestoreGhostedFieldArray(ctx.sFx);
	Field2D::RestoreGhostedFieldArray(ctx.sFy);

	//Offset pressure
	ForAllB(ctx.phi.n)
	{
		PetscScalar pMin;
		VecMin(ctx.pGas.g[b],NULL,&pMin);
		VecShift(ctx.pGas.g[b],-pMin);
		VecMax(ctx.pGas.g[b],NULL,&pMin);
		PetscPrintf(PETSC_COMM_WORLD,"  max(p) = %4.6f\n",pMin);
	}

	return(0);
}


PetscInt main( int argc, char* argv[] )
{
	PetscInitialize(&argc,&argv,(char*)0,help);
	Constant::ReadInputFile();

	PetscInt nb; PetscWrapper::ReadScalarInBinaryFormat(nb,"bubbleNumber");
	Context::Ctx ctx; Context::CreateContext(ctx,nb);
	Field2D::ScalarField pGasSum; Field2D::CreateField(ctx.da,pGasSum,"pGasSum");
	SNESWrapper::PetscNonLinearSolver pressureSolver; SNESWrapper::CreatePetscNonLinearSolver(ctx.da,pressureSolver,"pressureSolver");

	PetscInt tmin,tmax,dt;
	GetTimes(tmin,tmax,dt);
	for (PetscInt t=tmin; t<tmax; t+=dt)
	{
		Field2D::ReadFieldInBinaryFormat(ctx.phi,t);
		ComputeGasPressure(ctx,pressureSolver);
		Field2D::WriteFieldInBinaryFormat(ctx.pGas,t);
		Field2D::WriteFieldInVTSFormat(ctx.pGas,t);
		Field2D::SumVectorField(ctx.pGas,pGasSum);
		Field2D::WriteFieldInVTSFormat(pGasSum,t);
		Field2D::WriteFieldInBinaryFormat(pGasSum,t);
	}

	SNESWrapper::DestroyPetscNonLinearSolver(pressureSolver);
	Context::DestroyContext(ctx);
	Field2D::DestroyField(pGasSum);
	PetscFinalize();
	return(0);
}
