#ifndef FIELD2D_H
#define FIELD2D_H


#include "petsc.h"


/** \file
 *  \brief %Field functions conveniently grouping a global Petsc vector, a local Petsc vector and a multidimensional C-array */


/** \namespace Field
    \brief Collection of structures holding Petsc objects */
namespace Field2D
{


	/** \class ScalarField
	 *  \brief Structure holding a global PETSC vector, a local PETSC vector, and a 2-dimensional C-array */
	class ScalarField
	{


	public:


		DM da; /**< \brief Dynamic array */
		Vec g; /**< \brief Global Petsc vector */
		Vec l; /**< \brief Local Petsc vector */
		PetscScalar **a;/**< \brief 2-dimensional C-array, N=2*/


		/** \brief Overload parenthesis operator
		 *  \return reference to 2-dimensional C-array */
		PetscScalar **operator()()
		{
			return a;
		}


		/** \brief Overload parenthesis operator
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      reference to a[j][i] */
		 PetscScalar &operator()(PetscInt j,PetscInt i)
		{
			return a[j][i];
		}


	};


	/** \class VectorField
	 *  \brief Structure holding an array of global PETSC vectors, an array of local PETSC vectors, and an array of 2-dimensional C-arrays */
	class VectorField
	{


	public:


		DM da; /**< \brief Dynamic array */
		Vec *g; /**< \brief array of global Petsc vectors */
		Vec *l; /**< \brief array of local Petsc vectors */
		PetscScalar ***a; /**< \brief array of 2-dimensional C-arrays */
		PetscInt n; /**< \brief size */


		/** \brief Overload parenthesis operator
		 *  \return reference to array of 2-dimensional C-arrays a */
		PetscScalar ***operator()()
		{
			return a;
		}


		/** \brief Overload parenthesis operator
		 *  \param[in] dim space dimension
		 *  \return        reference to 2-dimensional C-array a[dim] */
		PetscScalar **operator()(PetscInt dim)
		{
			return a[dim];
		}


		/** \brief Overload parenthesis operator
		 *  \param[in] dim dimension
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        reference to C-array a[dim][j][i] */
		PetscScalar &operator()(PetscInt dim,PetscInt j,PetscInt i)
		{
			return a[dim][j][i];
		}


		/** \brief Get first C-array a[0]
		 *  \return reference to a[0] */
		PetscScalar **x()
		{
			return a[0];
		}


		/** \brief Get second C-array a[1]
		 *  \return reference to a[1] */
		PetscScalar **y()
		{
			return a[1];
		}


		/** \brief Get first C-array a[0][j][i]
		  * \param[in] j node index
		  * \param[in] i node index
		  * \return      reference to reference to C-array a[0][j][i] */
		PetscScalar &x(PetscInt j,PetscInt i)
		{
			return a[0][j][i];
		}


		/** \brief Get second C-array a[1][j][i]
		  * \param[in] j node index
		  * \param[in] i node index
		  * \return      reference to reference to C-array a[1][j][i] */
		 PetscScalar &y(PetscInt j, PetscInt i)
		{
			return a[1][j][i];
		}


	};


	/** \brief Constructor
	 *  \param[in] da   DMDA distributed array
	 *  \param[in] f    %ScalarField
	 *  \param[in] name global Petsc vector name */
	PetscErrorCode CreateField(DM da,ScalarField &f,const std::string& name);


	/** \brief Constructor
	 *  \param[in] da   DMDA distributed array
	 *  \param[in] F    %VectorField
	 *  \param[in] name global Petsc vector name
	 *  \param[in] n    size */
	PetscErrorCode CreateField(DM da,VectorField &F,const std::string& name,PetscInt n);


	/** \brief Two-dimensional constructor
	 *  \param[in] da   DMDA distributed array
	 *  \param[in] F    %VectorField
	 *  \param[in] name global Petsc vector name */
	PetscErrorCode Create2DField(DM da,VectorField &F,const std::string& name);


	/** \brief Destructor
	 *  \param[in] f %ScalarField */
	PetscErrorCode DestroyField(ScalarField &f);


	/** \brief Destructor
	 *  \param[in] F %VectorField */
	PetscErrorCode DestroyField(VectorField &F);


	/** \brief Get multidimensional C-array
	 *  \param[in] f %ScalarField */
	PetscErrorCode GetFieldArray(ScalarField &f);


	/** \brief Get array of multidimensional C-arrays
	 *  \param[in] F %VectorField */
	PetscErrorCode GetFieldArray(VectorField &F);


	/** \brief Get ONLY ONE multidimensional C-array
	 *  \param[in] F %VectorField
	 *  \param[in] i i-th element */
	PetscErrorCode GetFieldArray(VectorField &F,PetscInt i);


	/** \brief Restore multidimensional C-array
	 *  \param[in] f %ScalarField */
	PetscErrorCode RestoreFieldArray(ScalarField &f);


	/** \brief Restore array of multidimensional C-arrays
	 *  \param[in] F %VectorField */
	PetscErrorCode RestoreFieldArray(VectorField &F);


	/** \brief Restore ONLY ONE multidimensional C-array
	 *  \param[in] F  %VectorField
	 *  \param[in] i  i-th element */
	PetscErrorCode RestoreFieldArray(VectorField &F,PetscInt i);


	/** \brief Get multidimensional C-array containing ghost points
	 *  \param[in] f %ScalarField */
	PetscErrorCode GetGhostedFieldArray(ScalarField &f);


	/** \brief Get array of multidimensional C-arrays containing ghost points
	 *  \param[in] F %VectorField */
	PetscErrorCode GetGhostedFieldArray(VectorField &F);


	/** \brief Get ONLY ONE multidimensional C-array containing ghost points
	 *  \param[out] F %VectorField
	 *  \param[in]  i i-th element */
	PetscErrorCode GetGhostedFieldArray(VectorField &F,PetscInt i);


	/** \brief Restore multidimensional C-array containing ghost points
	 *  \param[in] f %ScalarField */
	PetscErrorCode RestoreGhostedFieldArray(ScalarField &f);


	/** \brief Restore array of multidimensional C-arrays containing ghost points
	 *  \param[in] F %VectorField */
	PetscErrorCode RestoreGhostedFieldArray(VectorField &F);


	/** \brief Restore ONLY ONE multidimensional C-array containing ghost points
	 *  \param[in] F %VectorField
	 *  \param[in] i i-th element */
	PetscErrorCode RestoreGhostedFieldArray(VectorField &F,PetscInt i);


	 /** \brief Write %ScalarField in binary format
	  *  \param[in] f %ScalarField
	  *  \param[in] t time index */
	PetscErrorCode WriteFieldInBinaryFormat(ScalarField &f,PetscInt t=-1);


	 /** \brief Write %VectorField in binary format
	  *  \param[in] F %VectorField
	  *  \param[in] t time index */
	PetscErrorCode WriteFieldInBinaryFormat(VectorField &F,PetscInt t=-1);


	 /** \brief Write %ScalarField in VTS format
	  *  \param[in] f %ScalarField
	  *  \param[in] t time index */
	PetscErrorCode WriteFieldInVTSFormat(ScalarField &f,PetscInt t=-1);


	 /** \brief Write %VectorField in VTS format
	  *  \param[in] F %VectorField
	  *  \param[in] t time index */
	PetscErrorCode WriteFieldInVTSFormat(VectorField &F,PetscInt t=-1);


	 /** \brief Read %ScalarField in binary format
	  *  \param[in] f %ScalarField
	  *  \param[in] t time index */
	PetscErrorCode ReadFieldInBinaryFormat(ScalarField &f,PetscInt t=-1);


	 /** \brief Read %VectorField in binary format
	  *  \param[in] F %VectorField
	  *  \param[in] t time index */
	PetscErrorCode ReadFieldInBinaryFormat(VectorField &F,PetscInt t=-1);


	/** \brief Set elements of the global Petsc vector to a given value
	  * \param[in] f %ScalarField
	  * \param[in] s scalar */
	PetscErrorCode SetField(ScalarField &f,PetscScalar s);


	/** \brief Set elements of the global Petsc vector to a given value
	  * \param[in] F %VectorField
	  * \param[in] s scalar */
	PetscErrorCode SetField(VectorField &F,PetscScalar s);


	/** \brief Set elements of the global Petsc vectors to a given value
	  * \param[in] F %VectorField
	  * \param[in] S one-dimensional scalar array S = [sa,sb,...] */
	PetscErrorCode SetField(VectorField &F,const PetscScalar *const S);


	/** \brief Compute sum of the global Petsc vector elements
	 *  \param[in] f %ScalarField
	 *  \param[in] s sum */
	PetscErrorCode ComputeSumOfFieldElements(ScalarField &f,PetscScalar &s);


	/** \brief Compute sum of global Petsc vector elements
	 *  \param[in]  F %VectorField
	 *  \param[out] S one-dimensional sum array */
	PetscErrorCode ComputeSumOfFieldElements(VectorField &F,PetscScalar *S);


	/** \brief Compute arithmetic mean of global Petsc vector elements
	 *  \param[in]  f %ScalarField f
	 *  \param[out] m arithmetic mean */
	PetscErrorCode ComputeArithmeticMeanOfFieldElements(ScalarField &f,PetscScalar &m);


	/** \brief Compute arithmetic mean of global Petsc vector elements
	 *  \param[in]  F %VectorField
	 *  \param[out] M one-dimensional arithmetic mean array */
	PetscErrorCode ComputeArithmeticMeanOfFieldElements(VectorField &F,PetscScalar *M);


	/** \brief Copy the global Petsc vector
	 *  \param[in]  from %ScalarField
	 *  \param[out] to   %ScalarField */
	PetscErrorCode CopyFieldTo(ScalarField &from,ScalarField &to);


	/** \brief Copy the array of global Petsc vectors
	 *  \param[in]  From VectorField
	 *  \param[out] To   VectorField */
	PetscErrorCode CopyFieldTo(VectorField &From,VectorField &To);


	/** \brief Sum elements of an array of global Petsc vectors copy to a %ScalarField
	 *  \param[in]  F %VectorField
	 *  \param[out] f %ScalarField */
	PetscErrorCode SumVectorField(VectorField &F,ScalarField &f);


	/** \brief Check whether two Petsc vectors are equal
	 *  \param[in] fa  %ScalarField
	 *  \param[in] fb  %ScalarField
	 *  \param[in] eps arbitrarily small value
	 *  \return true if |a-b| < epsilon */
	PetscBool AreFieldsEqual(ScalarField &fa,ScalarField &fb,PetscScalar eps=1e-8);


	/** \brief Check divergence
	 *  \param[in] f %ScalarField */
	PetscErrorCode CheckField(ScalarField &f);


	/** \brief Check divergence
	 *  \param[in] F %VectorField */
	PetscErrorCode CheckField(VectorField &F);


}


#endif
