#include "petsc.h"
#include "PetscWrapper.h"
#include "Field2D.h"


#include <iostream>
using namespace std;


static char help[] = "Interpolate field on two DAs\n\n";


PetscErrorCode GetTimes
(
	PetscInt &tmin,
	PetscInt &tmax,
	PetscInt &dt
)
{
	char str[PETSC_MAX_PATH_LEN];
	PetscBool flg;
	PetscOptionsGetString(NULL,NULL,"-t",str,PETSC_MAX_PATH_LEN,&flg);
	if (!flg)
	{
		SETERRQ(PETSC_COMM_WORLD,1,"Pass with -t tmin:tmax:dt");
	}
	char *p = strtok(str,":"); tmin = atoi(p);
	p = strtok(NULL,":"); tmax = (p == NULL) ? tmin+1 : atoi(p);
	p = strtok(NULL,":"); dt = (p == NULL) ? 1 : atoi(p);
	return(0);
}


PetscErrorCode CreateDMDA
(
	DM &da,
	PetscInt nx,
	PetscInt ny
)
{
	DMDACreate2d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		nx,                   //global dimension in x-direction of the array
		ny,                   //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		2,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		&da                   //dynamic array
	);
	DMSetUp(da);
	return(0);
}


PetscErrorCode DestroyDMDA
(
	DM &da
)
{
	DMDestroy(&da);
	return(0);
}


PetscErrorCode GetIndicesOfNodeClosestToP
(
	PetscScalar xp,
	PetscScalar yp,
	PetscWrapper::GridCoordinates &coords,
	PetscInt nx,
	PetscInt ny,
	PetscInt &i,
	PetscInt &j
)
{
	i=0;j=0;
	PetscScalar xMaxMinusDx = coords.x(0   ,nx-1);
	PetscScalar yMaxMinusDy = coords.y(ny-1,0   );
	PetscScalar xMax = nx;
	PetscScalar yMax = ny;
	if ( xMaxMinusDx < xp && xp <= xMax)
	{
		i=0;
	}
	else if ( xp>=0 && xp<=xMaxMinusDx )
	{
		i=0;
		while ( xp > coords.x(0  ,i+1) ){i++;}
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  xp out of bounds\n");
		PetscPrintf(PETSC_COMM_WORLD,"   xp = %4.3f\n",xp);
		PetscPrintf(PETSC_COMM_WORLD,"   bounds = [%4.3f,%4.3f]\n",0.,xMax);
		PetscEnd();
	}

	if ( yMaxMinusDy < yp && yp <= yMax)
	{
		j=0;
	}
	else if ( yp>=0 && yp<=yMaxMinusDy )
	{
		j=0;
		while ( yp > coords.y(j+1,0  ) ){j++;}
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  yp out of bounds\n");
		PetscPrintf(PETSC_COMM_WORLD,"   yp = %4.3f\n",yp);
		PetscPrintf(PETSC_COMM_WORLD,"   bounds = [%4.3f,%4.3f]\n",0.,yMax);
		PetscEnd();
	}
	return(0);
}


PetscScalar BilinearInterpolate
(
	PetscScalar xp,
	PetscScalar yp,
	PetscWrapper::GridCoordinates &coords,
	PetscInt nx,
	PetscInt ny,
 	PetscScalar **f
)
{
	//https://en.wikipedia.org/wiki/Bilinear_interpolation
	PetscInt i,j;
	GetIndicesOfNodeClosestToP(xp,yp,coords,nx,ny,i,j);
	PetscScalar x = xp-coords.x(j,i);
	PetscScalar y = yp-coords.y(j,i);
	PetscScalar f00 = f[j  ][i  ];
	PetscScalar f10 = f[j  ][i+1];
	PetscScalar f11 = f[j+1][i+1];
	PetscScalar f01 = f[j+1][i  ];
	return
	(
		  f00 * (1-x)*(1-y)
		+ f10 * (  x)*(1-y)
		+ f01 * (1-x)*(  y)
		+ f11 * (  x)*(  y)
	);
	return(0);
}



PetscInt main(int argc,char* argv[])
{
	PetscInitialize(&argc,&argv,(char*)0,help);
	PetscInt nxOld,nyOld,nxNew,nyNew;
	PetscBool flg;
	PetscOptionsGetInt(NULL,NULL,"-nxOld",&nxOld,&flg); if (!flg) SETERRQ(PETSC_COMM_WORLD,1," Pass -nxOld ??");
	PetscOptionsGetInt(NULL,NULL,"-nyOld",&nyOld,&flg); if (!flg) SETERRQ(PETSC_COMM_WORLD,1," Pass -nyOld ??");
	PetscOptionsGetInt(NULL,NULL,"-nxNew",&nxNew,&flg); if (!flg) SETERRQ(PETSC_COMM_WORLD,1," Pass -nxNew ??");
	PetscOptionsGetInt(NULL,NULL,"-nyNew",&nyNew,&flg); if (!flg) SETERRQ(PETSC_COMM_WORLD,1," Pass -nyNew ??");
	PetscScalar aspectRatioOld = static_cast<PetscScalar>(nyOld) / static_cast<PetscScalar>(nxOld);
	PetscScalar aspectRatioNew = static_cast<PetscScalar>(nyNew) / static_cast<PetscScalar>(nxNew);
	if ( aspectRatioOld != aspectRatioNew )
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Grid aspect ratio is not maintained\n");
		PetscEnd();
	}
	DM daOld; CreateDMDA(daOld,nxOld,nyOld); DMDASetUniformCoordinates(daOld,0,nxOld,0,nyOld,0,1);
	DM daNew; CreateDMDA(daNew,nxNew,nyNew); DMDASetUniformCoordinates(daNew,0,nxOld,0,nyOld,0,1);
	PetscWrapper::GridCoordinates coordsOld; PetscWrapper::GetPetscArray(daOld,coordsOld);
	PetscWrapper::GridCoordinates coordsNew; PetscWrapper::GetPetscArray(daNew,coordsNew);
	Field2D::VectorField fOld; Field2D::ScalarField fSumOld;
	Field2D::VectorField fNew; Field2D::ScalarField fSumNew;
	PetscInt nb; PetscOptionsGetInt(NULL,NULL,"-nb",&nb,&flg);
	if (!flg) SETERRQ(PETSC_COMM_WORLD,1," Pass -nb ??");
	char fieldNameOld[PETSC_MAX_PATH_LEN];
	char fieldNameNew[PETSC_MAX_PATH_LEN];
	PetscOptionsGetString(NULL,NULL,"-f",fieldNameOld,PETSC_MAX_PATH_LEN,&flg); if (!flg) SETERRQ(PETSC_COMM_WORLD,1,"  Pass field with the -f option");
	sprintf(fieldNameNew,"%sNew",fieldNameOld);
	Field2D::CreateField(daOld,fOld,fieldNameOld,nb); sprintf(fieldNameOld,"%sSum",fieldNameOld); Field2D::CreateField(daOld,fSumOld,fieldNameOld);
	Field2D::CreateField(daNew,fNew,fieldNameNew,nb); sprintf(fieldNameNew,"%sSum",fieldNameNew); Field2D::CreateField(daNew,fSumNew,fieldNameNew);
	PetscInt tMin,tMax,dt; GetTimes(tMin,tMax,dt);
	for (PetscInt t=tMin; t<tMax; t+=dt)
	{
		Field2D::ReadFieldInBinaryFormat(fOld,t);
		Field2D::GetFieldArray(fOld);
		Field2D::GetFieldArray(fNew);
		for (PetscInt jNew=0; jNew<nyNew; jNew++)
		{
			for (PetscInt iNew=0; iNew<nxNew; iNew++)
			{
				PetscScalar xp = coordsNew.x(jNew,iNew);
				PetscScalar yp = coordsNew.y(jNew,iNew);
				for (PetscInt b=0; b < nb; b++)
				{
					fNew(b,jNew,iNew) = BilinearInterpolate
					(
						xp,
						yp,
						coordsOld,
						nxOld,
						nyOld,
						fOld(b)
					);
// 					std::cout << "  f(" << xp << "," << yp << ") = " << fNew(b,jNew,iNew) << std::endl;
				}
			}
		}
		Field2D::RestoreFieldArray(fOld);
		Field2D::RestoreFieldArray(fNew);
		Field2D::WriteFieldInBinaryFormat(fNew,t);
		Field2D::SumVectorField(fOld,fSumOld);
		Field2D::SumVectorField(fNew,fSumNew);
		Field2D::WriteFieldInVTSFormat(fSumOld,t);
		Field2D::WriteFieldInVTSFormat(fSumNew,t);
	}
	PetscWrapper::RestorePetscArray(daOld,coordsOld);
	PetscWrapper::RestorePetscArray(daNew,coordsNew);
	Field2D::DestroyField(fOld);
	Field2D::DestroyField(fNew);
	DestroyDMDA(daOld);
	DestroyDMDA(daNew);
	PetscScalar alpha; 	PetscScalar *V0; PetscOptionsGetScalar(NULL,NULL,"-alpha",&alpha,&flg); if (!flg) SETERRQ(PETSC_COMM_WORLD,1," Pass -alpha ??");
	std::cout << "  Vnew/Vold = " << alpha << std::endl;
	PetscMalloc1(nb,&V0);
	PetscWrapper::ReadScalarInBinaryFormat(V0,"initialBubbleVolume",nb);
	for (PetscInt b=0; b < nb; b++)
	{
		V0[b] *= (alpha);
	}
	PetscWrapper::WriteScalarInBinaryFormat(V0,"initialBubbleVolumeNew",nb);
	PetscFree(V0);
	PetscFinalize();
}
