#include "Interpolation2D.h"


namespace
{


	const PetscInt X_ = 0; /**< \brief Space direction = 0 */
	const PetscInt Y_ = 1; /**< \brief Space direction = 1 */
	const PetscInt Z_ = 2; /**< \brief Space direction = 2 */
	const PetscInt C_ = 0; /**< \brief Centre index = 0 */
	const PetscInt N_ = 1; /**< \brief North index = 1 */
	const PetscInt S_ = 2; /**< \brief South index = 2 */
	const PetscInt E_ = 3; /**< \brief East index = 3 */
	const PetscInt W_ = 4; /**< \brief West index = 4 */
	const PetscInt F_ = 5; /**< \brief Front index = 5 */
	const PetscInt B_ = 6; /**< \brief Back index = 6 */
	const PetscScalar EDGE_INTERP_01_BY_02_ = 1./2.; /**< \brief Low-order edge interpolation coefficient = 1/2 */
	const PetscScalar EDGE_INTERP_09_BY_16_ = 9./16.; /**< \brief High-order edge interpolation coefficient = 9/16 */
	const PetscScalar EDGE_INTERP_01_BY_16_ = 1./16.; /**< \brief High-order edge interpolation coefficient = 1/16 */
	const PetscScalar FACE_INTERP_01_BY_04_ = 1./4.; /**< \brief Low-order face interpolation coefficient = 9/32 */
	const PetscScalar FACE_INTERP_09_BY_32_ = 9./32.; /**< \brief High-order face interpolation coefficient = 9/32 */
	const PetscScalar FACE_INTERP_01_BY_32_ = 1./32.; /**< \brief High-order face interpolation coefficient = 1/32 */


}


PetscScalar Interpolation2D::LowOrder::InterpolateXEdge
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		EDGE_INTERP_01_BY_02_ * ( f[j][i-1] + f[j][i] )
	);
}


PetscScalar Interpolation2D::LowOrder::InterpolateYEdge
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		EDGE_INTERP_01_BY_02_ * ( f[j-1][i] + f[j][i] )
	);
}


PetscScalar Interpolation2D::LowOrder::InterpolateEdge
(
	const PetscScalar *const *const f,
	PetscInt dim,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		Interpolation2D::LowOrder::InterpolateEdgePtr[dim](f,j,i)
	);
}


PetscScalar Interpolation2D::HighOrder::InterpolateXEdge
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  EDGE_INTERP_09_BY_16_ * ( f[j][i-1] + f[j][i  ] )
		- EDGE_INTERP_01_BY_16_ * ( f[j][i-2] + f[j][i+1] )
	);
}


PetscScalar Interpolation2D::HighOrder::InterpolateYEdge
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  EDGE_INTERP_09_BY_16_ * ( f[j-1][i] + f[j  ][i] )
		- EDGE_INTERP_01_BY_16_ * ( f[j-2][i] + f[j+1][i] )
	);
}


PetscScalar Interpolation2D::HighOrder::InterpolateEdge
(
	const PetscScalar *const *const f,
	PetscInt dim,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		Interpolation2D::HighOrder::InterpolateEdgePtr[dim](f,j,i)
	);
}


PetscScalar Interpolation2D::LowOrder::InterpolateXEdgeInverse
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		EDGE_INTERP_01_BY_02_ * ( 1./f[j][i-1] + 1./f[j][i] )
	);
}


PetscScalar Interpolation2D::LowOrder::InterpolateYEdgeInverse
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		EDGE_INTERP_01_BY_02_ * ( 1./f[j-1][i] + 1./f[j][i] )
	);
}


PetscScalar Interpolation2D::LowOrder::InterpolateEdgeInverse
(
	const PetscScalar *const *const f,
	PetscInt dim,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		Interpolation2D::LowOrder::InterpolateEdgeInversePtr[dim](f,j,i)
	);
}


PetscScalar Interpolation2D::HighOrder::InterpolateXEdgeInverse
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  EDGE_INTERP_09_BY_16_ * ( 1./f[j][i-1] + 1./f[j][i  ] )
		- EDGE_INTERP_01_BY_16_ * ( 1./f[j][i-2] + 1./f[j][i+1] )
	);
}


PetscScalar Interpolation2D::HighOrder::InterpolateYEdgeInverse
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  EDGE_INTERP_09_BY_16_ * ( 1./f[j-1][i] + 1./f[j  ][i] )
		- EDGE_INTERP_01_BY_16_ * ( 1./f[j-2][i] + 1./f[j+1][i] )
	);
}


PetscScalar Interpolation2D::HighOrder::InterpolateEdgeInverse
(
	const PetscScalar *const *const f,
	PetscInt dim,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		Interpolation2D::HighOrder::InterpolateEdgeInversePtr[dim](f,j,i)
	);
}


PetscScalar Interpolation2D::LowOrder::InterpolateFace
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		FACE_INTERP_01_BY_04_ *
		(
			  f[j-1][i-1]
			+ f[j-1][i  ]
			+ f[j  ][i-1]
			+ f[j  ][i  ]
		)
	);
}


PetscScalar Interpolation2D::HighOrder::InterpolateFace
(
	const PetscScalar *const *const f,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		FACE_INTERP_09_BY_32_*
		(
			  f[j-1][i-1]
			+ f[j-1][i  ]
			+ f[j  ][i-1]
			+ f[j  ][i  ]
		)

		- FACE_INTERP_01_BY_32_*
		(
			  f[j-2][i-2]
			+ f[j-2][i+1]
			+ f[j+1][i-2]
			+ f[j+1][i+1]
		)
	);
}


PetscErrorCode Interpolation2D::LowOrder::InterpolateXViscositiesAtCNSEW
(
	const PetscScalar *const *const eta,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[5]
)
{
	etaInterp[N_] = Interpolation2D::LowOrder::InterpolateFace(eta,j+1,i);
	etaInterp[S_] = Interpolation2D::LowOrder::InterpolateFace(eta,j  ,i);
	etaInterp[E_] = eta[j][i  ];
	etaInterp[W_] = eta[j][i-1];
	etaInterp[C_] = -( etaInterp[N_] + etaInterp[S_] + etaInterp[E_] + etaInterp[W_] );
	return(0);
}


PetscErrorCode Interpolation2D::LowOrder::InterpolateYViscositiesAtCNSEW
(
	const PetscScalar *const *const eta,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[5]
)
{
	etaInterp[N_] = eta[j  ][i];
	etaInterp[S_] = eta[j-1][i];
	etaInterp[E_] = Interpolation2D::LowOrder::InterpolateFace(eta,j,i+1);
	etaInterp[W_] = Interpolation2D::LowOrder::InterpolateFace(eta,j,i  );
	etaInterp[C_] = -( etaInterp[N_] + etaInterp[S_] + etaInterp[E_] + etaInterp[W_] );
	return(0);
}


PetscErrorCode Interpolation2D::LowOrder::InterpolateViscositiesAtCNSEW
(
	const PetscScalar *const *const eta,
	PetscInt dim,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[5]
)
{
	Interpolation2D::LowOrder::InterpolateViscositiesAtCNSEWPtr[dim](eta,j,i,etaInterp);
	return(0);
}


PetscErrorCode Interpolation2D::HighOrder::InterpolateXViscositiesAtCNSEW
(
	const PetscScalar *const *const eta,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[5]
)
{
	etaInterp[N_] = Interpolation2D::HighOrder::InterpolateFace(eta,j+1,i);
	etaInterp[S_] = Interpolation2D::HighOrder::InterpolateFace(eta,j  ,i);
	etaInterp[E_] = eta[j][i  ];
	etaInterp[W_] = eta[j][i-1];
	etaInterp[C_] = -( etaInterp[N_] + etaInterp[S_] + etaInterp[E_] + etaInterp[W_] );
	return(0);
}


PetscErrorCode Interpolation2D::HighOrder::InterpolateYViscositiesAtCNSEW
(
	const PetscScalar *const *const eta,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[5]
)
{
	etaInterp[N_] = eta[j  ][i];
	etaInterp[S_] = eta[j-1][i];
	etaInterp[E_] = Interpolation2D::HighOrder::InterpolateFace(eta,j,i+1);
	etaInterp[W_] = Interpolation2D::HighOrder::InterpolateFace(eta,j,i  );
	etaInterp[C_] = -( etaInterp[N_] + etaInterp[S_] + etaInterp[E_] + etaInterp[W_] );
	return(0);
}


PetscErrorCode Interpolation2D::HighOrder::InterpolateViscositiesAtCNSEW
(
	const PetscScalar *const *const eta,
	PetscInt dim,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[5]
)
{
	Interpolation2D::HighOrder::InterpolateViscositiesAtCNSEWPtr[dim](eta,j,i,etaInterp);
	return(0);
}



PetscErrorCode Interpolation2D::LowOrder::InterpolateInverseDensitiesAtCNSEW
(
	const PetscScalar *const *const rho,
	PetscInt j,
	PetscInt i,
	PetscScalar invRhoInterp[5]
)
{
	invRhoInterp[N_] = Interpolation2D::LowOrder::InterpolateYEdgeInverse(rho,j+1,i  );
	invRhoInterp[S_] = Interpolation2D::LowOrder::InterpolateYEdgeInverse(rho,j  ,i  );
	invRhoInterp[E_] = Interpolation2D::LowOrder::InterpolateXEdgeInverse(rho,j  ,i+1);
	invRhoInterp[W_] = Interpolation2D::LowOrder::InterpolateXEdgeInverse(rho,j  ,i  );
	invRhoInterp[C_] = -( invRhoInterp[N_] + invRhoInterp[S_] + invRhoInterp[E_] + invRhoInterp[W_] );
	return(0);
}


PetscErrorCode Interpolation2D::HighOrder::InterpolateInverseDensitiesAtCNSEW
(
	const PetscScalar *const *const rho,
	PetscInt j,
	PetscInt i,
	PetscScalar invRhoInterp[5]
)
{
	invRhoInterp[N_] = Interpolation2D::HighOrder::InterpolateYEdgeInverse(rho,j+1,i  );
	invRhoInterp[S_] = Interpolation2D::HighOrder::InterpolateYEdgeInverse(rho,j  ,i  );
	invRhoInterp[E_] = Interpolation2D::HighOrder::InterpolateXEdgeInverse(rho,j  ,i+1);
	invRhoInterp[W_] = Interpolation2D::HighOrder::InterpolateXEdgeInverse(rho,j  ,i  );
	invRhoInterp[C_] = -( invRhoInterp[N_] + invRhoInterp[S_] + invRhoInterp[E_] + invRhoInterp[W_] );
	return(0);
}
