#ifndef INTERPOLATION2D_H
#define INTERPOLATION2D_H


/** \file
 *  \brief Collection of interpolation schemes */


#include "petsc.h"


namespace
{


	typedef PetscScalar(*TwoDimensionalEdgeInterpolator)(const PetscScalar *const *const,PetscInt,PetscInt);
	typedef PetscErrorCode(*TwoDimensionalViscosityInterpolator)(const PetscScalar *const *const,PetscInt,PetscInt,PetscScalar*);


}


/** \namespace Interpolation2D
  * \brief Collection of interpolation schemes */
namespace Interpolation2D
{


	/** \namespace Interpolation2D::LowOrder
	 *  \brief Collection of low-order interpolation schemes */
	namespace LowOrder
	{


		/** \brief 2-point stencil edge interpolation along x-axis
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f|_{j,i-1/2} \f$ */
		PetscScalar InterpolateXEdge
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil edge interpolation along y-axis
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f|_{j-1/2,i} \f$ */
		PetscScalar InterpolateYEdge
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to 2-point-stencil edge interpolator */
		const TwoDimensionalEdgeInterpolator InterpolateEdgePtr[] =
		{
			Interpolation2D::LowOrder::InterpolateXEdge,
			Interpolation2D::LowOrder::InterpolateYEdge
		};


		/** \brief 2-point stencil edge interpolation
		 *  \param[in] f   two-dimensional array
		 *  \param[in] dim space dimension
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        interpolated scalar \f$ f|_{j    ,i-1/2} \f$ if dim = 0 else
		 *                                     \f$ f|_{j-1/2,i    } \f$ if dim = 1 */
		PetscScalar InterpolateEdge
		(
			const PetscScalar *const *const f,
			PetscInt dim,
			PetscInt j,
			PetscInt i
		);


		/** \brief Inverse 2-point stencil edge interpolation along x-axis
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ \left. \frac1{f} \right|_{j,i-1/2} \f$*/
		PetscScalar InterpolateXEdgeInverse
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Inverse 2-point stencil edge interpolation along y-axis
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ \left. \frac1{f} \right|_{j-1/2,i} \f$*/
		PetscScalar InterpolateYEdgeInverse
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to 2-point-stencil inverse edge interpolator */
		const TwoDimensionalEdgeInterpolator InterpolateEdgeInversePtr[] =
		{
			Interpolation2D::LowOrder::InterpolateXEdgeInverse,
			Interpolation2D::LowOrder::InterpolateYEdgeInverse
		};


		/** \brief Inverse 2-point stencil edge interpolation
		 *  \param[in] f   two-dimensional array
		 *  \param[in] dim space dimension
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        interpolated scalar \f$ \left. \frac1{f} \right|_{j    ,i-1/2} \f$ if dim = 0 else
		 *                                     \f$ \left. \frac1{f} \right|_{j-1/2,    i} \f$ if dim = 1 */
		PetscScalar InterpolateEdgeInverse
		(
			const PetscScalar *const *const f,
			PetscInt dim,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil face interpolation in z-plane
		 *  \param[in] f two-dimensional array
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f_{j-1/2,i-1/2} \f$ */
		PetscScalar InterpolateFace
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief One-dimensional array of interpolated viscosities needed in 5-point stencil Laplacian(eta,ux)
		 *  \note \f$ \eta_C = -(\eta_N + \eta_S + \eta_E + \eta_W) \f$
		 *  \note \f$ \eta_N = \eta_{ j+1/2, I     } = \eta_{ j+1/2, i-1/2 } \f$
		 *  \note \f$ \eta_S = \eta_{ j-1/2, I     } = \eta_{ j-1/2, i-1/2 } \f$
		 *  \note \f$ \eta_E = \eta_{ j    , I+1/2 } = \eta_{ j    , i     } \f$
		 *  \note \f$ \eta_W = \eta_{ j    , I-1/2 } = \eta_{ j    , i-1   } \f$
		 *  \param[in]  eta       two-dimensional array
		 *  \param[in]  j         node index
		 *  \param[in]  i         node index
		 *  \param[out] etaInterp viscosity array needed in \f$ \left. \nabla \cdot \left( \eta \nabla f_x \right ) \right|_{j,i-1/2} \f$ */
		PetscErrorCode InterpolateXViscositiesAtCNSEW
		(
			const PetscScalar *const *const eta,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[5]
		);


		/** \brief One-dimensional array of interpolated viscosities needed in 5-point stencil Laplacian(eta,uy)
		 *  \note \f$ \eta_C = -(\eta_N + \eta_S + \eta_E + \eta_W) \f$
		 *  \note \f$ \eta_N = \eta_{ J+1/2, i     } = \eta_{ j    , i     } \f$
		 *  \note \f$ \eta_S = \eta_{ J-1/2, i     } = \eta_{ j-1  , i     } \f$
		 *  \note \f$ \eta_E = \eta_{ J    , i+1/2 } = \eta_{ j-1/2, i+1/2 } \f$
		 *  \note \f$ \eta_W = \eta_{ J    , i-1/2 } = \eta_{ j-1/2, i-1/2 } \f$
		 *  \param[in]  eta       two-dimensional array
		 *  \param[in]  j         node index
		 *  \param[in]  i         node index
		 *  \param[out] etaInterp viscosity array needed in \f$ \left. \nabla \cdot \left( \eta \nabla f_y \right ) \right|_{j-1/2,i} \f$ */
		PetscErrorCode InterpolateYViscositiesAtCNSEW
		(
			const PetscScalar *const *const eta,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[5]
		);


		/** \brief Pointer to 5-point viscosity interpolator needed in Laplacian(eta,f) */
		const TwoDimensionalViscosityInterpolator InterpolateViscositiesAtCNSEWPtr[] =
		{
			Interpolation2D::LowOrder::InterpolateXViscositiesAtCNSEW,
			Interpolation2D::LowOrder::InterpolateYViscositiesAtCNSEW
		};


		/** brief One-dimensional array of interpolated viscosities needed in 5-point stencil Laplacian(eta,u[dim])
		 *  \param[in]  eta       two-dimensional array
		 *  \param[in]  dim       space dimension
		 *  \param[in]  j         node index
		 *  \param[in]  i         node index
		 *  \param[out] etaInterp viscosity array needed in \f$ \nabla \cdot \left( \eta \nabla f_{\alpha=0,1} \right ) \f$
		                          @ \f$ |_{j    ,i-1/2} \f$ if dim = 0 else
		                          @ \f$ |_{j-1/2,    i} \f$ if dim = 1  */
		PetscErrorCode InterpolateViscositiesAtCNSEW
		(
			const PetscScalar *const *const eta,
			PetscInt dim,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[5]
		);


		/** \brief One-dimensional array of interpolated inverted densities needed in 5-point stencil Div(Grad(p)/rho)
		 *  \note \f$ \left. \frac1{\rho} \right|_C = -(\left. \frac1{\rho} \right|_N + \left. \frac1{\rho} \right|_S + \left. \frac1{\rho} \right|_E + \left. \frac1{\rho} \right|_W) \f$
		 *  \note \f$ \left. \frac1{\rho} \right|_N = \left. \frac1{\rho} \right|_{ j+1/2, i     } \f$
		 *  \note \f$ \left. \frac1{\rho} \right|_S = \left. \frac1{\rho} \right|_{ j-1/2, j     } \f$
		 *  \note \f$ \left. \frac1{\rho} \right|_E = \left. \frac1{\rho} \right|_{ j    , i+1/2 } \f$
		 *  \note \f$ \left. \frac1{\rho} \right|_W = \left. \frac1{\rho} \right|_{ j    , i-1/2 } \f$
		 *  \param[in]  rho          two-dimensional array
		 *  \param[in]  j            node index
		 *  \param[in]  i            node index
		 *  \param[out] invRhoInterp one-dimensional array used in \f$ \left. \nabla \cdot \left( \frac{\nabla p}{\rho} \right ) \right|_{j,i} \f$ */
		PetscErrorCode InterpolateInverseDensitiesAtCNSEW
		(
			const PetscScalar *const *const rho,
			PetscInt j,
			PetscInt i,
			PetscScalar invRhoInterp[5]
		);


	}


	/** \namespace Interpolation2D::HighOrder
	 *  \brief Collection of high-order interpolation schemes */
	namespace HighOrder
	{


		/** \brief 4-point stencil edge interpolation along x-axis
		 *  \copydetails Interpolation2D::LowOrder::InterpolateXEdge */
		PetscScalar InterpolateXEdge
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil edge interpolation along y-axis
		 *  \copydetails Interpolation2D::LowOrder::InterpolateYEdge */
		PetscScalar InterpolateYEdge
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to 4-point-stencil edge interpolator */
		const TwoDimensionalEdgeInterpolator InterpolateEdgePtr[] =
		{
			Interpolation2D::HighOrder::InterpolateXEdge,
			Interpolation2D::HighOrder::InterpolateYEdge
		};

		/** \brief 4-point stencil edge interpolation
		 *  \copydetails Interpolation2D::LowOrder::InterpolateEdge */
		PetscScalar InterpolateEdge
		(
			const PetscScalar *const *const f,
			PetscInt dim,
			PetscInt j,
			PetscInt i
		);


		/** \brief Inverse 4-point stencil edge interpolation along x-axis
		 *  \copydetails Interpolation2D::LowOrder::InterpolateXEdgeInverse */
		PetscScalar InterpolateXEdgeInverse
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Inverse 4-point stencil edge interpolation along y-axis
		 *  \copydetails Interpolation2D::LowOrder::InterpolateYEdgeInverse */
		PetscScalar InterpolateYEdgeInverse
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to 4-point-stencil inverse edge interpolator */
		const TwoDimensionalEdgeInterpolator InterpolateEdgeInversePtr[] =
		{
			Interpolation2D::HighOrder::InterpolateXEdgeInverse,
			Interpolation2D::HighOrder::InterpolateYEdgeInverse
		};


		/** \brief Inverse 4-point stencil edge interpolation
		 *  \copydetails Interpolation2D::LowOrder::InterpolateEdgeInverse */
		PetscScalar InterpolateEdgeInverse
		(
			const PetscScalar *const *const f,
			PetscInt dim,
			PetscInt j,
			PetscInt i
		);


		/** \brief 8-point stencil face interpolation in z-plane
		 *  \copydetails Interpolation2D::LowOrder::InterpolateFace */
		PetscScalar InterpolateFace
		(
			const PetscScalar *const *const f,
			PetscInt j,
			PetscInt i
		);


		/** \copydoc Interpolation2D::LowOrder::InterpolateXViscositiesAtCNSEW */
		PetscErrorCode InterpolateXViscositiesAtCNSEW
		(
			const PetscScalar *const *const eta,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[5]
		);


		/** \copydoc Interpolation2D::LowOrder::InterpolateYViscositiesAtCNSEW */
		PetscErrorCode InterpolateYViscositiesAtCNSEW
		(
			const PetscScalar *const *const eta,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[5]
		);


		/** \brief Pointer to 5-point viscosity interpolator needed in Laplacian(eta,f) */
		const TwoDimensionalViscosityInterpolator InterpolateViscositiesAtCNSEWPtr[] =
		{
			Interpolation2D::HighOrder::InterpolateXViscositiesAtCNSEW,
			Interpolation2D::HighOrder::InterpolateYViscositiesAtCNSEW
		};


		/** \copydoc Interpolation2D::LowOrder::InterpolateViscositiesAtCNSEW */
		PetscErrorCode InterpolateViscositiesAtCNSEW
		(
			const PetscScalar *const *const eta,
			PetscInt dim,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[5]
		);


		/** \copydoc Interpolation2D::LowOrder::InterpolateInverseDensitiesAtCNSEW */
		PetscErrorCode InterpolateInverseDensitiesAtCNSEW
		(
			const PetscScalar *const *const rho,
			PetscInt j,
			PetscInt i,
			PetscScalar invRhoInterp[5]
		);


	}


}


#endif
