#include "Constant.h"
#include "Context.h"
#include "Interpolation2D.h"
#include "Discretisation2D.h"
#include "PhiEquation.h"
#include "PressureEquation.h"
#include "UEquation.h"
#include "SNESWrapper.h"
#include "Utils.h"
#include "PetscWrapper.h"


using namespace Context;
using namespace Field2D;
using namespace SNESWrapper;


static char help[] = "Foam dynamics.\n\n";


namespace
{
	const PetscInt X_ = 0;
	const PetscInt Y_ = 1;
}


PetscErrorCode UpdatePhi
(
	Ctx &ctx,
	PetscNonLinearSolver &phiSolver,
	PetscBool expTimeIntegration
)
{
	PhiEquation::ComputeBulk(ctx);
	PhiEquation::ComputeRepulsion(ctx);
	PhiEquation::ComputeRHS(ctx);
	if ( expTimeIntegration )
	{
		CopyFieldTo(ctx.phiRHS,ctx.phi);
	}
	else
	{
		for (PetscInt b=0; b<(ctx.phi.n); ++b)
		{
			ctx.b = b;
			SetNonLinearSolverSolution(phiSolver,ctx.phi.g[b]);
			RunPetscNonLinearSolver
			(
				phiSolver,
				PhiEquation::FormSNESFunction,
				PhiEquation::FormSNESJacobian,
				&ctx
			);
			GetNonLinearSolverSolution(phiSolver,ctx.phi.g[b]);
		}
	}
	return(0);
}


PetscErrorCode UpdateU
(
	Ctx &ctx,
	PetscNonLinearSolver &pressureSolver
)
{
	PhiEquation::ComputeSurfaceForce(ctx);
	UEquation::ComputeEtaRho(ctx);
	UEquation::ComputeRHS(ctx);
	UEquation::ComputeDivRHS(ctx);
	RunPetscNonLinearSolver
	(
		pressureSolver,
		PressureEquation::FormSNESFunction,
		PressureEquation::FormSNESJacobian,
		&ctx
	);
	GetNonLinearSolverSolution(pressureSolver,ctx.p.g);
	UEquation::SubstractGradP(ctx);
	CopyFieldTo(ctx.uRHS,ctx.u);
	UEquation::SubstractMomentumExcess(ctx);
	return(0);
}


PetscErrorCode ComputeGasPressure
(
	Ctx &ctx,
	PetscNonLinearSolver &pressureSolver
)
{
	//Compute bubble volume
	GetFieldArray(ctx.phi);
	ForAllB(ctx.phi.n)
	{
		PetscScalar lV = 0.;
		PetscScalar gV;
		ForAllJI(ctx.da)
		{
			lV += ctx.phi(b,j,i);
		}
		MPI_Allreduce(&lV,&gV,1,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
		PetscPrintf(PETSC_COMM_WORLD,"  V[%i] = %4.2f \n",b,gV);
	}
	RestoreFieldArray(ctx.phi);

	//Compute surface force
	PhiEquation::ComputeSurfaceForce(ctx);

	//Loop over each bubble
	GetGhostedFieldArray(ctx.sFx);
	GetGhostedFieldArray(ctx.sFy);
	ForAllB(ctx.phi.n)
	{
		//Compute Div(surface force)
		PetscScalar alpha = (3. * PetscSqrtScalar(2.) ) * (ctx.xi) * (ctx.invDx3);
		GetFieldArray(ctx.divRHS);
		ForAllJI(ctx.da)
		{
			ctx.divRHS(j,i) =
			(
				Discretisation2D::SecondOrder::Divergent(ctx.sFx(b),ctx.sFy(b),j,i) * alpha
			); //(1/dx) moved to pressure equation
		}
		RestoreFieldArray(ctx.divRHS);

		//Compute gas pressure
		SetNonLinearSolverSolution(pressureSolver,ctx.pGas.g[b]);
		RunPetscNonLinearSolver
		(
			pressureSolver,
			PressureEquation::FormSNESFunction,
			PressureEquation::FormSNESJacobian,
			&ctx
		);
		GetNonLinearSolverSolution(pressureSolver,ctx.pGas.g[b]);
	}
	RestoreGhostedFieldArray(ctx.sFx);
	RestoreGhostedFieldArray(ctx.sFy);

	//Offset pressure
	ForAllB(ctx.phi.n)
	{
		PetscScalar pMin;
		VecMin(ctx.pGas.g[b],NULL,&pMin);
		VecShift(ctx.pGas.g[b],-pMin);
		VecMax(ctx.pGas.g[b],NULL,&pMin);
		PetscPrintf(PETSC_COMM_WORLD,"  max(p) = %4.6f\n",pMin);
	}

	return(0);
}


PetscInt main(int argc,char* argv[])
{
	//Initialize Petsc
	PetscInitialize(&argc,&argv,(char*)0,help);

	//Read global constants from input file
	Constant::ReadInputFile();

	//Solution context
	Ctx ctx;
	if ( (Constant::Time::TBeginEq)  > -1 && (Constant::Time::TBegin) == -1 )
	{
		PetscInt nb; PetscWrapper::ReadScalarInBinaryFormat(nb,"bubbleNumber");
		PetscPrintf(PETSC_COMM_WORLD,"  Restart equilibrium with nb = %i\n",nb);
		CreateContext(ctx,nb);
		ReadContext(ctx,Constant::Time::TBeginEq);
		PhiEquation::MakeFoam(ctx);
		WriteFieldInVTSFormat(ctx.phiSum,Constant::Time::TBeginEq);
	}
	else if ( (Constant::Time::TBeginEq) == -1 && (Constant::Time::TBegin)  > -1 )
	{
		PetscInt nb; PetscWrapper::ReadScalarInBinaryFormat(nb,"bubbleNumber");
		PetscPrintf(PETSC_COMM_WORLD,"  Restart unsteady with nb = %i\n",nb);
		CreateContext(ctx,nb);
		ReadContext(ctx,Constant::Time::TBegin);
		PhiEquation::MakeFoam(ctx);
		WriteFieldInVTSFormat(ctx.phiSum,Constant::Time::TBegin);
	}
	else if ( (Constant::Time::TBeginEq) == -1 && (Constant::Time::TBegin) == -1 )
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Single bubble\n");
		PetscInt nb = 1;
		CreateContext(ctx,nb);
		InitFoamToSingleBubble(ctx);
		PhiEquation::MakeFoam(ctx);
		WriteContext(ctx);
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Exit - iterations were not set correctly - \n");
		PetscEnd();
	}

	//SNES solvers
	PetscNonLinearSolver phiSolver;
	PetscNonLinearSolver pressureSolver;
	CreatePetscNonLinearSolver(ctx.da,phiSolver,"phiSolver");
	CreatePetscNonLinearSolver(ctx.da,pressureSolver,"pressureSolver");

	//Time integration flag
	const PetscBool expTimeIntegration = (Constant::Numerical::ThetaPhi) < 0.001 ? PETSC_TRUE : PETSC_FALSE;

	//Equilibrium
	PetscInt n = Constant::Time::TEndEq - Constant::Time::TBeginEq;
	if ( n > 0)
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Equilibrium\n");

		//Time loop
		for (PetscInt t=0; t<n; ++t)
		{
			if ( t % (Constant::Time::PrintEveryEq) == 0 || t == (n-1) )
			{
				PetscPrintf(PETSC_COMM_WORLD,"  Iteration = %i\n",t);
			}

			UpdatePhi
			(
				ctx,
				phiSolver,
				expTimeIntegration
			);

			if ( ( t > 0 && t % (Constant::Time::WriteEveryEq) == 0 ) || t == (n-1) )
			{
// 				ComputeGasPressure(ctx,pressureSolver);
				PhiEquation::MakeFoam(ctx);
				CheckField(ctx.phiSum);
				WriteContext(ctx);
			}
		}
	}

	//Unsteady
	n = Constant::Time::TEnd - Constant::Time::TBegin;
	if ( n > 0)
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Unsteady\n");

		GetFieldArray(ctx.u);
		GetGhostedFieldArray(ctx.phiWall);
		ForAllJI(ctx.da)
		{
			PetscScalar zetax = 1. - Interpolation2D::LowOrder::InterpolateEdge(ctx.phiWall(),X_,j,i); //phi @ (j    ,i-1/2)
			PetscScalar zetay = 1. - Interpolation2D::LowOrder::InterpolateEdge(ctx.phiWall(),Y_,j,i); //phi @ (j-1/2,i    )
			PetscScalar ux = Constant::Foam::UFoam[X_];
			PetscScalar uy = Constant::Foam::UFoam[Y_];
			ctx.u.x(j,i) = zetax * ux;
			ctx.u.y(j,i) = zetay * uy;
		}
		RestoreFieldArray(ctx.u);
		RestoreGhostedFieldArray(ctx.phiWall);

		//Time loop
		for (PetscInt t=0; t<n; ++t)
		{
			if ( t % (Constant::Time::PrintEvery) == 0 || t == (n-1) )
			{
				PetscPrintf(PETSC_COMM_WORLD,"  Iteration = %i\n",t);
			}

			UpdatePhi
			(
				ctx,
				phiSolver,
				expTimeIntegration
			);

// 			UpdateU
// 			(
// 				ctx,
// 				pressureSolver
// 			);

			if ( ( t > 0 && t % (Constant::Time::WriteEvery) == 0 ) || t == (n-1) )
			{
// 				ComputeGasPressure(ctx,pressureSolver);
				PhiEquation::MakeFoam(ctx);
				CheckField(ctx.phiSum);
				WriteContext(ctx);
			}
		}
	}

	//Destroy
	DestroyContext(ctx);
	DestroyPetscNonLinearSolver(phiSolver);
	DestroyPetscNonLinearSolver(pressureSolver);
	PetscFinalize();
	return(0);
}
