#include "PetscWrapper.h"


namespace
{
	const PetscInt WIDTH_ = 4; /**< \brief Number of time digits in file name set using ".*" */
}


PetscErrorCode PetscWrapper::GetPetscArray
(
	DM da,
	GridCoordinates &gc
)
{
	DMGetCoordinateDM(da,&gc.cda);
	DMGetCoordinates(da,&gc.g);
	DMDAVecGetArrayRead(gc.cda,gc.g,&gc.a);
	return(0);
}


PetscErrorCode PetscWrapper::RestorePetscArray
(
	DM da,
	GridCoordinates &gc
)
{
	DMDAVecRestoreArrayRead(gc.cda,gc.g,&gc.a);
	return(0);
}


PetscErrorCode PetscWrapper::CreateGlobalPetscVector
(
	DM da,
	Vec &g,
	const std::string& name
)
{
	DMCreateGlobalVector(da,&g);
	VecSet(g,0.0);
	PetscObjectSetName((PetscObject)g,name.c_str());
	PetscPrintf(PETSC_COMM_WORLD,"  Global Petsc vector %s created \n", name.c_str());
	return(0);
}


PetscErrorCode PetscWrapper::CreateGlobalPetscVector
(
	DM da,
	Vec *&g,
	const std::string& name,
	PetscInt n
)
{
	Vec vtmp;
	DMCreateGlobalVector(da,&vtmp);
	VecDuplicateVecs(vtmp,n,&g);
	VecDestroy(&vtmp);
	for (PetscInt i=0;i<n;++i)
	{
		VecSet(g[i],0.);
		PetscObjectSetName((PetscObject)g[i],name.c_str());
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Array of %i global Petsc vectors %s created\n",n,name.c_str());
	return(0);
}


PetscErrorCode PetscWrapper::CreateLocalPetscVector
(
	DM da,
	Vec &l
)
{
	DMCreateLocalVector(da,&l);
	PetscPrintf(PETSC_COMM_WORLD,"  Local Petsc vector created\n");
	return(0);
}


PetscErrorCode PetscWrapper::CreateLocalPetscVector
(
	DM da,
	Vec *&l,
	PetscInt n
)
{
	Vec vtmp;
	DMCreateLocalVector(da,&vtmp);
	VecDuplicateVecs(vtmp,n,&l);
	VecDestroy(&vtmp);
	PetscPrintf(PETSC_COMM_WORLD,"  Array of %i local Petsc vectors created\n",n);
	return(0);
}


PetscErrorCode PetscWrapper::DestroyPetscVector
(
	Vec &v
)
{
	VecDestroy(&v);
	PetscPrintf(PETSC_COMM_WORLD,"  Petsc vector destroyed\n");
	return(0);
}


PetscErrorCode PetscWrapper::DestroyPetscVector
(
	Vec *&v,
	PetscInt n
)
{
	VecDestroyVecs(n,&v);
	PetscPrintf(PETSC_COMM_WORLD,"  Array of %i Petsc vectors destroyed\n",n);
	return(0);
}


PetscErrorCode PetscWrapper::GetLocalPetscVector
(
	DM da,
	Vec &l
)
{
	DMGetLocalVector(da,&l);
	return(0);
}


PetscErrorCode PetscWrapper::GetLocalPetscVector
(
	DM da,
	Vec *&l,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMGetLocalVector(da,&l[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::RestoreLocalPetscVector
(
	DM da,
	Vec &l
)
{
	DMRestoreLocalVector(da,&l);
	return(0);
}


PetscErrorCode PetscWrapper::RestoreLocalPetscVector
(
	DM da,
	Vec *&l,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMRestoreLocalVector(da,&l[i]);
	}
	return(0);
}


template<typename Type>
PetscErrorCode PetscWrapper::GetPetscArray
(
	DM da,
	Vec g,
	Type &a
)
{
	DMDAVecGetArray(da,g,&a);
	return(0);
}


template<typename Type>
PetscErrorCode PetscWrapper::GetPetscArray
(
	DM da,
	Vec *g,
	Type *&a,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMDAVecGetArray(da,g[i],&a[i]);
	}
	return(0);
}


template<typename Type>
PetscErrorCode PetscWrapper::RestorePetscArray
(
	DM da,
	Vec g,
	Type &a
)
{
	DMDAVecRestoreArray(da,g,&a);
	return(0);
}


template<typename Type>
PetscErrorCode PetscWrapper::RestorePetscArray
(
	DM da,
	Vec *g,
	Type *&a,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMDAVecRestoreArray(da,g[i],&a[i]);
	}
	return(0);
}


template<typename Type>
PetscErrorCode PetscWrapper::GetGhostedPetscArray
(
	DM da,
	Vec g,
	Vec l,
	Type &a
)
{
	DMGlobalToLocalBegin(da,g,INSERT_VALUES,l);
	DMGlobalToLocalEnd(da,g,INSERT_VALUES,l);
	DMDAVecGetArrayRead(da,l,&a);
	return(0);
}


template<typename Type>
PetscErrorCode PetscWrapper::GetGhostedPetscArray
(
	DM da,
	Vec *g,
	Vec *l,
	Type *&a,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMGlobalToLocalBegin(da,g[i],INSERT_VALUES,l[i]);
		DMGlobalToLocalEnd(da,g[i],INSERT_VALUES,l[i]);
		DMDAVecGetArrayRead(da,l[i],&a[i]);
	}
	return(0);
}


template<typename Type>
PetscErrorCode PetscWrapper::RestoreGhostedPetscArray
(
	DM da,
	Vec l,
	Type &a
)
{
	DMDAVecRestoreArrayRead(da,l,&a);
	return(0);
}


template<typename Type>
PetscErrorCode PetscWrapper::RestoreGhostedPetscArray
(
	DM da,
	Vec *l,
	Type *&a,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMDAVecRestoreArrayRead(da,l[i],&a[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::WritePetscVectorInBinaryFormat
(
	Vec g,
	PetscInt t
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g,&vname);
	char fname[PETSC_MAX_PATH_LEN];
	if (t<0)
	{
		sprintf(fname,"%s.dat",vname);
	}
	else
	{
		sprintf(fname,"%s_%.*d.dat",vname,WIDTH_,t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	VecView(g,viewer);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WritePetscVectorInBinaryFormat
(
	Vec *g,
	PetscInt t,
	PetscInt n
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g[0],&vname);
	char fname[PETSC_MAX_PATH_LEN];
	if (t<0)
	{
		sprintf(fname,"%s.dat",vname);
	}
	else
	{
		sprintf(fname,"%s_%.*d.dat",vname,WIDTH_,t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	for (PetscInt i=0;i<n;++i)
	{
		char prefix[1000];
		sprintf(prefix,"%03d",i);
		PetscObjectSetOptionsPrefix((PetscObject)g[i],prefix);
		VecView(g[i],viewer);
	}
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WritePetscVectorInVTSFormat
(
	Vec g,
	PetscInt t
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g,&vname);
	char fname[PETSC_MAX_PATH_LEN];
	if (t<0)
	{
		sprintf(fname,"%s.vts",vname);
	}
	else
	{
		sprintf(fname,"%s_%.*d.vts",vname,WIDTH_,t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerVTKOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	VecView(g,viewer);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WritePetscVectorInVTSFormat
(
	Vec *g,
	PetscInt t,
	PetscInt n
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g[0],&vname);
	for(PetscInt i=0;i<n;++i)
	{
		char fname[PETSC_MAX_PATH_LEN];
		if (t<0)
		{
			sprintf(fname,"%s%.*d.vts",vname,WIDTH_,i);
		}
		else
		{
			sprintf(fname,"%s%.*d_%.*d.vts",vname,WIDTH_,i,WIDTH_,t);
		}
		PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
		PetscViewer viewer;
		PetscViewerVTKOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
		VecView(g[i],viewer);
		PetscViewerDestroy(&viewer);
	}
	return(0);
}


PetscErrorCode PetscWrapper::ReadPetscVectorInBinaryFormat
(
	Vec g,
	PetscInt t
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g,&vname);
	char fname[PETSC_MAX_PATH_LEN];
	if (t<0)
	{
		sprintf(fname,"%s.dat",vname);
	}
	else
	{
		sprintf(fname,"%s_%.*d.dat",vname,WIDTH_,t);
	}
	PetscBool fileExists;
	PetscTestFile(fname,'\0',&fileExists);
	if (!fileExists)
	{
		char errorMessage[PETSC_MAX_PATH_LEN];
		sprintf(errorMessage,"  Error - %s not found\n",fname);
		PetscPrintf(PETSC_COMM_WORLD,errorMessage);
		PetscEnd();
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	VecLoad(g,viewer);
	PetscViewerDestroy(&viewer);
	return(0);
}



PetscErrorCode PetscWrapper::ReadPetscVectorInBinaryFormat
(
	Vec *g,
	PetscInt t,
	PetscInt n
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g[0],&vname);
	char fname[PETSC_MAX_PATH_LEN];
	if (t<0)
	{
		sprintf(fname,"%s.dat",vname);
	}
	else
	{
		sprintf(fname,"%s_%.*d.dat",vname,WIDTH_,t);
	}
	PetscBool fileExists;
	PetscTestFile(fname,'\0',&fileExists);
	if (!fileExists)
	{
		char errorMessage[PETSC_MAX_PATH_LEN];
		sprintf(errorMessage,"  Error - %s not found\n",fname);
		PetscPrintf(PETSC_COMM_WORLD,errorMessage);
		PetscEnd();
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	for (PetscInt i=0;i<n;++i)
	{
		char prefix[1000];
		sprintf(prefix,"%03d",i);
		PetscObjectSetOptionsPrefix((PetscObject)g[i],prefix);
		VecLoad(g[i],viewer);
	}
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInBinaryFormat
(
	PetscInt s,
	const std::string &name
)
{
	char fname[PETSC_MAX_PATH_LEN];
	sprintf(fname,"%s.dat",name.c_str());
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	#if PETSC_VERSION_GT(3,13,2)
		PetscViewerBinaryWrite(viewer,&s,1,PETSC_INT);
	#else
		PetscViewerBinaryWrite(viewer,&s,1,PETSC_INT,PETSC_FALSE);
	#endif
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInBinaryFormat
(
	PetscInt *s,
	const std::string &name,
	PetscInt n
)
{
	char fname[PETSC_MAX_PATH_LEN];
	sprintf(fname,"%s.dat",name.c_str());
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	#if PETSC_VERSION_GT(3,13,2)
		PetscViewerBinaryWrite(viewer,s,n,PETSC_INT);
	#else
		PetscViewerBinaryWrite(viewer,s,n,PETSC_INT,PETSC_FALSE);
	#endif
	PetscViewerDestroy(&viewer);
	return(0);
}



PetscErrorCode PetscWrapper::WriteScalarInBinaryFormat
(
	PetscScalar s,
	const std::string &name
)
{
	char fname[PETSC_MAX_PATH_LEN];
	sprintf(fname,"%s.dat",name.c_str());
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	#if PETSC_VERSION_GT(3,13,2)
		PetscViewerBinaryWrite(viewer,&s,1,PETSC_SCALAR);
	#else
		PetscViewerBinaryWrite(viewer,&s,1,PETSC_SCALAR,PETSC_FALSE);
	#endif
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInBinaryFormat
(
	PetscScalar *s,
	const std::string &name,
	PetscInt n
)
{
	char fname[PETSC_MAX_PATH_LEN];
	sprintf(fname,"%s.dat",name.c_str());
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	#if PETSC_VERSION_GT(3,13,2)
		PetscViewerBinaryWrite(viewer,s,n,PETSC_SCALAR);
	#else
		PetscViewerBinaryWrite(viewer,s,n,PETSC_SCALAR,PETSC_FALSE);
	#endif
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::ReadScalarInBinaryFormat
(
	PetscInt &s,
	const std::string &name
)
{
	char fname[PETSC_MAX_PATH_LEN];
	sprintf(fname,"%s.dat",name.c_str());
	PetscBool fileExists;
	PetscTestFile(fname,'\0',&fileExists);
	if (!fileExists)
	{
		char errorMessage[PETSC_MAX_PATH_LEN];
		sprintf(errorMessage,"  Error - %s not found\n",fname);
		PetscPrintf(PETSC_COMM_WORLD,errorMessage);
		PetscEnd();
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,&s,1,NULL,PETSC_INT);
	PetscViewerDestroy(&viewer);
	MPI_Bcast(&s,1,MPIU_INT,0,PETSC_COMM_WORLD);
	return(0);
}


PetscErrorCode PetscWrapper::ReadScalarInBinaryFormat
(
	PetscScalar &s,
	const std::string &name
)
{
	char fname[PETSC_MAX_PATH_LEN];
	sprintf(fname,"%s.dat",name.c_str());
	PetscBool fileExists;
	PetscTestFile(fname,'\0',&fileExists);
	if (!fileExists)
	{
		char errorMessage[PETSC_MAX_PATH_LEN];
		sprintf(errorMessage,"  Error - %s not found\n",fname);
		PetscPrintf(PETSC_COMM_WORLD,errorMessage);
		PetscEnd();
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,&s,1,NULL,PETSC_SCALAR);
	PetscViewerDestroy(&viewer);
	MPI_Bcast(&s,1,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	return(0);
}


PetscErrorCode PetscWrapper::ReadScalarInBinaryFormat
(
	PetscInt *s,
	const std::string &name,
	PetscInt n
)
{
	char fname[PETSC_MAX_PATH_LEN];
	sprintf(fname,"%s.dat",name.c_str());
	PetscBool fileExists;
	PetscTestFile(fname,'\0',&fileExists);
	if (!fileExists)
	{
		char errorMessage[PETSC_MAX_PATH_LEN];
		sprintf(errorMessage,"  Error - %s not found\n",fname);
		PetscPrintf(PETSC_COMM_WORLD,errorMessage);
		PetscEnd();
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,s,n,NULL,PETSC_INT);
	PetscViewerDestroy(&viewer);
	MPI_Bcast(s,n,MPIU_INT,0,PETSC_COMM_WORLD);
	return(0);
}


PetscErrorCode PetscWrapper::ReadScalarInBinaryFormat
(
	PetscScalar *s,
	const std::string &name,
	PetscInt n
)
{
	char fname[PETSC_MAX_PATH_LEN];
	sprintf(fname,"%s.dat",name.c_str());
	PetscBool fileExists;
	PetscTestFile(fname,'\0',&fileExists);
	if (!fileExists)
	{
		char errorMessage[PETSC_MAX_PATH_LEN];
		sprintf(errorMessage,"  Error - %s not found\n",fname);
		PetscPrintf(PETSC_COMM_WORLD,errorMessage);
		PetscEnd();
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,s,n,NULL,PETSC_SCALAR);
	PetscViewerDestroy(&viewer);
	MPI_Bcast(s,n,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	return(0);
}


PetscErrorCode PetscWrapper::SetPetscVector
(
	Vec g,
	PetscScalar s
)
{
	VecSet(g,s);
	return(0);
}


PetscErrorCode PetscWrapper::SetPetscVector
(
	Vec *g,
	PetscScalar s,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecSet(g[i],s);
	}
	return(0);
}


PetscErrorCode PetscWrapper::SetPetscVector
(
	Vec *g,
	const PetscScalar *const s,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecSet(g[i],s[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::ComputeSumOfPetscVectorElements
(
	Vec v,
	PetscScalar &s
)
{
	VecSum(v,&s);
	return(0);
}


PetscErrorCode PetscWrapper::ComputeSumOfPetscVectorElements
(
	Vec *v,
	PetscScalar *s,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecSum(v[i],&s[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::ComputeArithmeticMeanOfPetscVectorElements
(
	Vec g,
	PetscScalar &m
)
{
	PetscInt size;
	VecGetSize(g,&size);
	VecSum(g,&m);
	m /= size;
	return(0);
}


PetscErrorCode PetscWrapper::ComputeArithmeticMeanOfPetscVectorElements
(
	Vec *g,
	PetscScalar *m,
	PetscInt n
)
{
	PetscInt size;
	for (PetscInt i=0;i<n;++i)
	{
		VecGetSize(g[i],&size);
		VecSum(g[i],&m[i]);
		m[i] /= size;
	}
	return(0);
}


// PetscErrorCode PetscWrapper::CopyPetscVectorToAndRaiseElements
// (
// 	Vec a,
// 	PetscScalar exp,
// 	Vec b
// )
// {
// 	VecCopy(a,b);
// 	VecPow(b,exp);
// 	return(0);
// }


// PetscErrorCode PetscWrapper::CopyPetscVectorToAndRaiseElements
// (
// 	Vec *a,
// 	PetscScalar exp,
// 	Vec *b,
// 	PetscInt n
// )
// {
// 	for (PetscInt i=0;i<n;++i)
// 	{
// 		VecCopy(a[i],b[i]);
// 		VecPow(b[i],exp);
// 	}
// 	return(0);
// }


PetscErrorCode PetscWrapper::CopyPetscVectorTo
(
	Vec from,
	Vec to
)
{
	VecCopy(from,to);
	return(0);
}


PetscErrorCode PetscWrapper::CopyPetscVectorTo
(
	Vec *from,
	Vec *to,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecCopy(from[i],to[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::SumArrayOfPetscVectors
(
	Vec *g,
	Vec gSum,
	PetscInt n
)
{
	PetscScalar *a_gSum; VecGetArray(gSum,&a_gSum);
	PetscInt localN; VecGetLocalSize(gSum,&localN);
	const PetscScalar *a_g;
	PetscInt v=0;
	{
		//g0
		VecGetArrayRead(g[v],&a_g);
		for (PetscInt i=0; i<localN; i++)
		{
			a_gSum[i] = a_g[i];
		}
		VecRestoreArrayRead(g[v],&a_g);
	}
	for (v=1; v<n; ++v)
	{
		//g1,g2,...gn-1
		VecGetArrayRead(g[v],&a_g);
		for (PetscInt i=0; i<localN; i++)
		{
			a_gSum[i] += a_g[i];
		}
		VecRestoreArrayRead(g[v],&a_g);
	}
	VecRestoreArray(gSum,&a_gSum);
	return(0);
}


PetscBool PetscWrapper::ArePetscVectorsEqual
(
	Vec va,
	Vec vb,
	PetscScalar eps
)
{
	Vec w;
	VecDuplicate(va,&w);
	VecWAXPY(w,-1,va,vb);
	PetscScalar norm;
	VecNorm(w,NORM_2,&norm);
	VecDestroy(&w);
	//PetscPrintf(PETSC_COMM_WORLD,"  Norm( = %4.6f\n",norm);
	if ( norm < eps )
	{
		return PETSC_TRUE;
	}
	return PETSC_FALSE;
}


PetscErrorCode PetscWrapper::CheckPetscVector
(
	Vec g
)
{
	PetscScalar min; VecMin(g,NULL,&min);
	PetscScalar max; VecMax(g,NULL,&max);

	max = PetscMax
	(
		PetscAbsReal(min),
		PetscAbsReal(max)
	);
	PetscPrintf(PETSC_COMM_WORLD,"  Max = %4.4f\n",max);

	if ( PetscIsInfOrNanScalar(max) || max > 10000. )
	{
		const char *name;
		PetscObjectGetName((PetscObject)g,&name);
		std::string msg = "  Error - " + std::string(name) + " -> infinity";
		PetscPrintf(PETSC_COMM_WORLD,"%s\n",msg.c_str());
		PetscEnd();
	}
	return(0);
}


PetscErrorCode PetscWrapper::CheckPetscVector
(
	Vec *g,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		PetscWrapper::CheckPetscVector(g[i]);
	}
	return(0);
}


//2D explicit instanciation
template PetscErrorCode PetscWrapper::GetPetscArray<PetscScalar**>(DM,Vec,PetscScalar**&);
template PetscErrorCode PetscWrapper::GetPetscArray<PetscScalar**>(DM,Vec*,PetscScalar***&,PetscInt);
template PetscErrorCode PetscWrapper::RestorePetscArray<PetscScalar**>(DM,Vec,PetscScalar**&);
template PetscErrorCode PetscWrapper::RestorePetscArray<PetscScalar**>(DM,Vec*,PetscScalar***&,PetscInt);
template PetscErrorCode PetscWrapper::GetGhostedPetscArray<PetscScalar**>(DM,Vec,Vec,PetscScalar**&);
template PetscErrorCode PetscWrapper::GetGhostedPetscArray<PetscScalar**>(DM,Vec*,Vec*,PetscScalar***&,PetscInt);
template PetscErrorCode PetscWrapper::RestoreGhostedPetscArray<PetscScalar**>(DM,Vec,PetscScalar**&a);
template PetscErrorCode PetscWrapper::RestoreGhostedPetscArray<PetscScalar**>(DM,Vec*,PetscScalar***&a,PetscInt);


//3D explicit instanciation
template PetscErrorCode PetscWrapper::GetPetscArray<PetscScalar***>(DM,Vec ,PetscScalar***&);
template PetscErrorCode PetscWrapper::GetPetscArray<PetscScalar***>(DM,Vec*,PetscScalar****&,PetscInt);
template PetscErrorCode PetscWrapper::RestorePetscArray<PetscScalar***>(DM,Vec ,PetscScalar***&);
template PetscErrorCode PetscWrapper::RestorePetscArray<PetscScalar***>(DM,Vec*,PetscScalar****&,PetscInt);
template PetscErrorCode PetscWrapper::GetGhostedPetscArray<PetscScalar***>(DM,Vec ,Vec ,PetscScalar***&);
template PetscErrorCode PetscWrapper::GetGhostedPetscArray<PetscScalar***>(DM,Vec*,Vec*,PetscScalar****&,PetscInt);
template PetscErrorCode PetscWrapper::RestoreGhostedPetscArray<PetscScalar***>(DM,Vec ,PetscScalar***&a);
template PetscErrorCode PetscWrapper::RestoreGhostedPetscArray<PetscScalar***>(DM,Vec*,PetscScalar****&a,PetscInt);
