#ifndef PETSCWRAPPER_H
#define PETSCWRAPPER_H


#include "petsc.h"


/** \file
 *  \brief Collection of Petsc functions */


/** \namespace PetscWrapper
  * \brief Collection of Petsc functions */
namespace PetscWrapper
{


	/** \class %GridCoordinates
	 *  \brief Object for rapid and easy access to DM grid coordinates */
	class GridCoordinates
	{
	public:
		Vec g;               /**< \brief global Petsc vector holding coordinates */
		DM  cda;             /**< \brief Petsc coordinate DM */
		const DMDACoor2d **a;/**< \brief Petsc structure holding 2d coordinates. */


		/** \brief Const reference to x[j][i]
		*   \param[in] j node index
		*   \param[in] i node index
		*   \return const reference to a[j][i].x */
		const PetscScalar& x(PetscInt j,PetscInt i){return a[j][i].x;}


		/** \brief Const reference to y[j][i]
		*   \param[in] j node index
		*   \param[in] i node index
		*   \return const reference to a[j][i].y */
		const PetscScalar& y(PetscInt j,PetscInt i){return a[j][i].y;}
	};

	/** \brief Get Petsc coordinate array
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] gc Grid coordinate object */
	PetscErrorCode GetPetscArray(DM da,GridCoordinates &gc);


	/** \brief Restore coordinate array
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] gc Grid coordinate object */
	PetscErrorCode RestorePetscArray(DM da,GridCoordinates &gc);


	/** \brief Create global Petsc vector
	 *  \param[in]  da   DMDA distributed array
	 *  \param[out] g    global Petsc vector
	 *  \param[in]  name Petsc vector name */
	PetscErrorCode CreateGlobalPetscVector(DM da,Vec &g,const std::string &name);


	/** \brief Create array of global Petsc vectors
	 *  \param[in]  da   DMDA distributed array
	 *  \param[out] g    array of global Petsc vectors g = [g0,g1,...]
	 *  \param[in]  name Petsc vector name
	 *  \param[in]  n    size */
	PetscErrorCode CreateGlobalPetscVector(DM da,Vec *&g,const std::string &name,PetscInt n);


	/** \brief Create local Petsc vector with extra space for ghost values
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  local vector */
	PetscErrorCode CreateLocalPetscVector(DM da,Vec &l);


	/** \brief Create array of local Petsc vectors with extra space for ghost values
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  array of local Petsc vectors l = [l0,l1,...]
	 *  \param[in]  n  size */
	PetscErrorCode CreateLocalPetscVector(DM da,Vec *&l,PetscInt n);


	/** \brief Destroy Petsc vector
	 *  \param[out] v global/local vector */
	PetscErrorCode DestroyPetscVector(Vec &v);


	/** \brief Destroy array of Petsc vectors
	 *  \param[out] v array of global/local vector  v = [v0,v1,...]
	 *  \param[in]  n size */
	PetscErrorCode DestroyPetscVector(Vec *&v,PetscInt n);


	/** \brief Get local Petsc vector with extra space for ghost values
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  local vector */
	PetscErrorCode GetLocalPetscVector(DM da,Vec &l);


	/** \brief Get array of local Petsc vectors with extra space for ghost values
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  array of local Petsc vectors l = [l0,l1,...]
	 *  \param[in]  n  size */
	PetscErrorCode GetLocalPetscVector(DM da,Vec *&l,PetscInt n);


	/** \brief Restore local Petsc vector
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  local vector */
	PetscErrorCode RestoreLocalPetscVector(DM da,Vec &l);


	/** \brief Restore array of local Petsc vectors
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  array of local Petsc vectors l = [l0,l1,...]
	 *  \param[in]  n  size */
	PetscErrorCode RestoreLocalPetscVector(DM da,Vec *&,PetscInt n);


	/** \brief Get multidimensional C-array from global Petsc vector
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  global vector
	 *  \param[out] a  multidimensional C-array */
	template<typename Type>
	PetscErrorCode GetPetscArray(DM da,Vec g,Type &a);


	/** \brief Get array of multidimensional C-arrays from array of global Petsc vectors
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  array of global Petsc vectors g = [g0,g1,...]
	 *  \param[out] a  array of multidimensional C-arrays a = [a0,a0,...]
	 *  \param[in]  n  size */
	template<typename Type>
	PetscErrorCode GetPetscArray(DM da,Vec *g,Type *&a,PetscInt n);


	/** \brief Restore global Petsc vector from multidimensional C-array
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  global vector
	 *  \param[out] a  multidimensional C-array */
	template<typename Type>
	PetscErrorCode RestorePetscArray(DM da,Vec g,Type &a);


	/** \brief Restore array of global Petsc vectors from array of multidimensional C-arrays
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  array of global Petsc vectors g = [g0,g1,...]
	 *  \param[out] a  array of multidimensional C-arrays a = [a0,a1,...]
	 *  \param[in]  n  size */
	template<typename Type>
	PetscErrorCode RestorePetscArray(DM da,Vec *g,Type *&a,PetscInt n);


	/** \brief Get multidimensional C-array containing ghost points. This is done after scattering global Petsc vector to local parts
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  global vector
	 *  \param[out] l  local vector
	 *  \param[out] a  multidimensional C-array */
	template<typename Type>
	PetscErrorCode GetGhostedPetscArray(DM da,Vec g,Vec l,Type &a);


	/** \brief Get array of multidimensional C-arrays containing ghost points. This is done after scattering global Petsc vectors to local parts
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  array of global Petsc vectors g = [g0,g1,...]
	 *  \param[out] l  array of local Petsc vectors l = [l0,l1,...]
	 *  \param[out] a  array of multidimensional C-arrays a = [a0,a1,...]
	 *  \param[in]  n  size */
	template<typename Type>
	PetscErrorCode GetGhostedPetscArray(DM da,Vec *g,Vec *l,Type *&a,PetscInt n);


	/** \brief Restore local Petsc vector from multidimensional C-array containing ghost points
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  l  local vector
	 *  \param[out] a  multidimensional C-array */
	template<typename Type>
	PetscErrorCode RestoreGhostedPetscArray(DM da,Vec l,Type &a);


	/** \brief Restore array of local Petsc vectors from array of multidimensional C-arrays containing ghost points
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  l  array of local Petsc vectors l = [l0,l1,...]
	 *  \param[out] a  array of multidimensional C-arrays a = [a0,a1,...]
	 *  \param[in]  n  size */
	template<typename Type>
	PetscErrorCode RestoreGhostedPetscArray(DM da,Vec *l,Type *&a,PetscInt n);


	/** \brief Write Petsc vector in binary format
	 *  \param[in] g global Petsc vector
	 *  \param[in] t time index */
	PetscErrorCode WritePetscVectorInBinaryFormat(Vec g,PetscInt t);


	/** \brief Write array of Petsc vectors in binary format
	 *  \param[in] g array of global Petsc vector g = [g0,g1,...]
	 *  \param[in] t time index
	 *  \param[in] n size */
	PetscErrorCode WritePetscVectorInBinaryFormat(Vec *g,PetscInt t,PetscInt n);


	/** \brief Read Petsc vector in binary format
	 *  \param[out] g global Petsc vector
	 *  \param[in]  t time index */
	PetscErrorCode ReadPetscVectorInBinaryFormat(Vec g,PetscInt t);


	/** \brief Read array of Petsc vectors in binary format
	 *  \param[out] g array of global Petsc vectors g = [g0,g1,...]
	 *  \param[in]  t time index
	 *  \param[in]  n size */
	PetscErrorCode ReadPetscVectorInBinaryFormat(Vec *g,PetscInt t,PetscInt n);


	/** \brief Write Petsc vector in VTS format
	 *  \param[in] g global Petsc vector
	 *  \param[in] t time index */
	PetscErrorCode WritePetscVectorInVTSFormat(Vec g,PetscInt t);


	/** \brief Write array of Petsc vectors in VTS format
	 *  \param[in] g array of global Petsc vector g = [g0,g1,...]
	 *  \param[in] t time index
	 *  \param[in] n size */
	PetscErrorCode WritePetscVectorInVTSFormat(Vec *g,PetscInt t,PetscInt n);


	/** \brief Write integer
	 *  \param[in] s    integer
	 *  \param[in] name file name */
	PetscErrorCode WriteScalarInBinaryFormat(PetscInt s,const std::string &name);


	/** \brief Write one-dimensional integer array
	 *  \param[in] s    integer
	 *  \param[in] name file name
	 *  \param[in] n    size */
	PetscErrorCode WriteScalarInBinaryFormat(PetscInt *s,const std::string &name,PetscInt n);


	/** \brief Write scalar
	 *  \param[in] s    scalar
	 *  \param[in] name file name */
	PetscErrorCode WriteScalarInBinaryFormat(PetscScalar s,const std::string &name);


	/** \brief Write one-dimensional scalar array
	 *  \param[in] s    one-dimensional array
	 *  \param[in] name file name
	 *  \param[in] n    size */
	PetscErrorCode WriteScalarInBinaryFormat(PetscScalar *s,const std::string &name,PetscInt n);


	/** \brief Read integer
	 *  \param[in] s    integer
	 *  \param[in] name file name */
	PetscErrorCode ReadScalarInBinaryFormat(PetscInt &s,const std::string &name);


	/** \brief Read one-dimensional integer array
	 *  \param[in] s    integer
	 *  \param[in] name file name
	 *  \param[in] n    size */
	PetscErrorCode ReadScalarInBinaryFormat(PetscInt *s,const std::string &name,PetscInt n);


	/** \brief Read scalar
	 *  \param[in] s    scalar
	 *  \param[in] name file name */
	PetscErrorCode ReadScalarInBinaryFormat(PetscScalar &s,const std::string &name);


	/** \brief Read one-dimensional scalar array
	 *  \param[in] s    one-dimensional scalar array
	 *  \param[in] name file name
	 *  \param[in] n    size */
	PetscErrorCode ReadScalarInBinaryFormat(PetscScalar *s,const std::string &name,PetscInt n);


	/** \brief Set vector elements to a scalar value
	 *  \param[out] g global Petsc vector
	 *  \param[in]  s scalar */
	PetscErrorCode SetPetscVector(Vec g,PetscScalar s);


	/** \brief Set vector elements to a scalar value
	 *  \param[out] g array of global Petsc vectors g = [g0,g1,...]
	 *  \param[in]  s scalar
	 *  \param[in]  n size */
	PetscErrorCode SetPetscVector(Vec *g,PetscScalar s,PetscInt n);


	/** \brief Set vector elements to a scalar value
	 *  \param[out] g array of global Petsc vector g = [g0,g1,...]
	 *  \param[in]  s scalar array s = [s0,s1,...]
	 *  \param[in]  n size*/
	PetscErrorCode SetPetscVector(Vec *g,const PetscScalar *const s,PetscInt n);


	/** \brief Sum elements of a global Petsc vector
	*   \param[in]  g global Petsc vector
	*   \param[out] s sum */
	PetscErrorCode ComputeSumOfPetscVectorElements(Vec g, PetscScalar &s);


	/** \brief Sum elements of each Petsc vector
	*   \param[in]  g array of global Petsc vectors g = [g0,g1,...]
	*   \param[out] s sum array s = [s0,s1,...]
	*   \param[in]  n size */
	PetscErrorCode ComputeSumOfPetscVectorElements(Vec *g, PetscScalar *s,PetscInt n);


	/** \brief Arithmetic mean of elements in Petsc vector
	 *  \param[in]  g global Petsc vector
	 *  \param[out] m mean */
	PetscErrorCode ComputeArithmeticMeanOfPetscVectorElements(Vec g,PetscScalar &m);


	/** \brief Arithmetic mean of elements in Petsc vector
	 *  \param[in]  g Array of Petsc vectors g = [g0,g1,...]
	 *  \param[out] m array of means m = [m0,m1,...]
	 *  \param[in]  n size*/
	PetscErrorCode ComputeArithmeticMeanOfPetscVectorElements(Vec *g,PetscScalar *m,PetscInt n);


// 	/** \brief Copy and raise each element of a global Petsc vector
// 	 *  \param[in]  a   Global Petsc vector
// 	 *  \param[in]  exp Exponent
// 	 *  \param[out] b   Global Petsc vector */
// 	PetscErrorCode CopyPetscVectorToAndRaiseElements(Vec a,PetscScalar exp,Vec b);


// 	/** \brief Copy and raise each element of a global Petsc vector
// 	 *  \param[in]  a   Array of global Petsc vectors
// 	 *  \param[in]  exp Exponent
// 	 *  \param[out] b   Array of global Petsc vector */
// 	PetscErrorCode CopyPetscVectorToAndRaiseElements(Vec *a,PetscScalar exp,Vec *b,PetscInt n);


	/** \brief Copy global Petsc vector
	 *  \param[in]  from global Petsc vector
	 *  \param[out] to   global Petsc vector */
	PetscErrorCode CopyPetscVectorTo(Vec from,Vec to);


	/** \brief Copy array of global Petsc vectors
	*   \param[in]  from array of Petsc vectors from = [from0,from1,...]
	*   \param[out] to   array of Petsc vectors to = [to0,to1,...]
	*   \param[in]  n    size */
	PetscErrorCode CopyPetscVectorTo(Vec *from,Vec *to,PetscInt n);


	/** \brief Addition an array of Petsc vectors
	 *  \param[in]  g    array of Petsc vectors g = [g0,g1,...]
	 *  \param[out] gSum resulting sum as global Petsc vector
	 *  \param[in]  n    size */
	PetscErrorCode SumArrayOfPetscVectors(Vec *g,Vec gSum,PetscInt n);


	/** \brief Check whether two Petsc vectors are equal
	 *  \param[in] a  global Petsc vector
	 *  \param[in] b  global Petsc vector
	 *  \param[in] eps return true if |va-vb| < epsilon */
	PetscBool ArePetscVectorsEqual(Vec a,Vec b,PetscScalar eps=1e-8);


	/** \brief Check convergence of a global Petsc vector
	 *  \param[in] g global Petsc vector */
	PetscErrorCode CheckPetscVector(Vec g);


	/** \brief Check convergence of a array of Petsc vectors
	 *  \param[in] g array of global Petsc vector g = [g0,g1,...]
	 *  \param[in] n size */
	PetscErrorCode CheckPetscVector(Vec *g,PetscInt n);


}


#endif


/*
	Read-only access is done as follows
	const PetscScalar ***a;
	DMDAVecGetPetscArrayRead(da,g,&a);
	... = a[k][j][i];
	DMDAVecRestorePetscArrayRead(da,g,&a);
*/
