#include "PhiEquation.h"
#include "Discretisation2D.h"
#include "Interpolation2D.h"
#include "PetscWrapper.h"
#include "Utils.h"


using namespace PetscWrapper;
using namespace Field2D;
using namespace Interpolation2D;
using namespace Discretisation2D;


namespace
{
	const PetscInt X_ = 0; /**< \brief Space direction = 0 */
	const PetscInt Y_ = 1; /**< \brief Space direction = 1 */
}


PetscErrorCode PhiEquation::ComputeBubbleVolume
(
	Context::Ctx &ctx
)
{
	GetFieldArray(ctx.phi);
	ForAllB(ctx.phi.n)
	{
		PetscScalar lV = 0.;
		PetscScalar gV;
		ForAllJI(ctx.da)
		{
			lV += ctx.phi(b,j,i);
		}
		MPI_Allreduce(&lV,&gV,1,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
		ctx.bubbleVolume[b] = gV;
	}
	RestoreFieldArray(ctx.phi);
	return(0);
}


PetscErrorCode PhiEquation::ComputeSqrBubbleVolume
(
	Context::Ctx &ctx
)
{
	GetFieldArray(ctx.phi);
	ForAllB(ctx.phi.n)
	{
		PetscScalar lV = 0.;
		PetscScalar gV;
		ForAllJI(ctx.da)
		{
			lV += PetscSqr( ctx.phi(b,j,i) );
		}
		MPI_Allreduce(&lV,&gV,1,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
		ctx.bubbleVolume[b] = gV;
	}
	RestoreFieldArray(ctx.phi);
	return(0);
}


PetscErrorCode PhiEquation::ComputeBulk
(
	Context::Ctx &ctx
)
{
// 	ComputeBubbleVolume(ctx);    // Mueller, PRL, 2019
	ComputeSqrBubbleVolume(ctx); // Lober, SR, 2015
	GetFieldArray(ctx.phi);
	GetFieldArray(ctx.muBulk);
	ForAllB(ctx.phi.n)
	{
		//corr < 0 [growth] else [shrinkage]
		PetscScalar corr = (ctx.bubbleVolume[b])/(ctx.initialBubbleVolume[b]) - 1.;
		PetscScalar phic = 0.5 + (ctx.bubbleGrowthCoefficient) * corr;
		ForAllJI(ctx.da)
		{
			PetscScalar phi = ctx.phi(b,j,i);
			ctx.muBulk(b,j,i) = 4. * ( phi - 1. ) * ( phi - phic ) *  phi;
		}
// 		PetscPrintf(PETSC_COMM_WORLD,"  phi_correction = %4.2f\n",corr);
	}
	RestoreFieldArray(ctx.phi);
	RestoreFieldArray(ctx.muBulk);
	return(0);
}


PetscErrorCode PhiEquation::ComputeRepulsion
(
	Context::Ctx &ctx
)
{
	const PetscScalar threshold = 1e-4;
	auto Pow2 = [&threshold](const PetscScalar& x)
	{
		//Truncated lambda function Pow2(x) = x^2
		return x > threshold ? PetscSqr(x) : 0. ;
	};
	GetFieldArray(ctx.phiWall);
	GetFieldArray(ctx.phi);
	GetFieldArray(ctx.muRep);
	for (PetscInt p=0; p<(ctx.phi.n); ++p)
	{
		ForAllJI(ctx.da)
		{
			PetscScalar phip = ctx.phi(p,j,i);
			PetscScalar phiqSqr = Pow2( ctx.phiWall(j,i) );
			for(PetscInt q=0; q<(ctx.phi.n); ++q)
			{
				if ( q != p )
				{
					phiqSqr += Pow2( ctx.phi(q,j,i) );
				}
			}
			ctx.muRep(p,j,i) = (ctx.repulsionCoefficient) * phip * phiqSqr ;
		}
	}
	RestoreFieldArray(ctx.phiWall);
	RestoreFieldArray(ctx.phi);
	RestoreFieldArray(ctx.muRep);
	return(0);
}


PetscErrorCode PhiEquation::ComputeRHS
(
	Context::Ctx &ctx
)
{
	GetFieldArray(ctx.phiRHS);
	GetGhostedFieldArray(ctx.u);
	GetGhostedFieldArray(ctx.phi);
	GetGhostedFieldArray(ctx.muBulk);
	GetGhostedFieldArray(ctx.muRep);
	ForAllB(ctx.phi.n)
	{
		ForAllJI(ctx.da)
		{
			PetscScalar uDir[2] =
			{
				LowOrder::InterpolateEdge(ctx.u.x(),X_,j  ,i+1), //ux @ (j,i) = (j    ,I+1/2)
				LowOrder::InterpolateEdge(ctx.u.y(),Y_,j+1,i  )  //uy @ (j,i) = (J+1/2,i    )
			};
			PetscScalar uGradPhi = FirstOderUpwind::UGrad(uDir,ctx.phi(b),j,i);
			PetscScalar lapPhi = FourthOrder::Laplacian(ctx.phi(b),j,i);
			ctx.phiRHS(b,j,i) =
			(
				  ctx.phi(b,j,i)
				- (ctx.dtInvDx1) * uGradPhi
				- (ctx.dtInvPe) * ctx.muBulk(b,j,i)
				- (ctx.dtInvPe) * ctx.muRep(b,j,i)
				+ (ctx.xi2DtInvPeExp) * lapPhi
			);
		}
	}
	RestoreFieldArray(ctx.phiRHS);
	RestoreGhostedFieldArray(ctx.u);
	RestoreGhostedFieldArray(ctx.phi);
	RestoreGhostedFieldArray(ctx.muBulk);
	RestoreGhostedFieldArray(ctx.muRep);
	return(0);
}


PetscErrorCode PhiEquation::ComputeSurfaceForce
(
	Context::Ctx &ctx
)
{
	GetFieldArray(ctx.sFx);
	GetFieldArray(ctx.sFy);
	GetGhostedFieldArray(ctx.phi);
	ForAllB(ctx.phi.n)
	{
		//nx,ny @(j-1/2,i-1/2) = (J,I)
		{
			GetFieldArray(ctx.nv);
			ForAllJI(ctx.da)
			{
				ctx.nv.x(j,i) = 0.5 * ( ctx.phi(b,j  ,i  ) + ctx.phi(b,j-1,i  ) )
				              - 0.5 * ( ctx.phi(b,j  ,i-1) + ctx.phi(b,j-1,i-1) );

				ctx.nv.y(j,i) = 0.5 * ( ctx.phi(b,j  ,i  ) + ctx.phi(b,j  ,i-1) )
				              - 0.5 * ( ctx.phi(b,j-1,i  ) + ctx.phi(b,j-1,i-1) );
			}
			RestoreFieldArray(ctx.nv);
		}

		//Temporary surface force @(j,i)
		{
			GetFieldArray(ctx.sFTmp);
			GetGhostedFieldArray(ctx.nv);
			ForAllJI(ctx.da)
			{
				PetscScalar nx,ny,n2,nmag;
				PetscScalar epsilon = 1e-10;
				PetscScalar kappa = 0;

				//Curvature @(j+1/2,i+1/2) = (J+1,I+1)
				nx = ctx.nv.x(j+1,i+1);
				ny = ctx.nv.y(j+1,i+1);
				n2 = PetscSqr(nx) + PetscSqr(ny);
				if (n2 > epsilon)
				{
					kappa += ( + nx + ny ) / PetscSqrtReal( n2 );
				}

				//Curvature @(j-1/2,i+1/2) = (J,I+1)
				nx = ctx.nv.x(j,i+1);
				ny = ctx.nv.y(j,i+1);
				n2 = PetscSqr(nx) + PetscSqr(ny);
				if (n2 > epsilon)
				{
					kappa += ( + nx - ny ) / PetscSqrtReal( n2 );
				}

				//Curvature @(j+1/2,i-1/2) = (J+1,I)
				nx = ctx.nv.x(j+1,i);
				ny = ctx.nv.y(j+1,i);
				n2 = PetscSqr(nx) + PetscSqr(ny);
				if (n2 > epsilon)
				{
					kappa += ( - nx + ny ) / PetscSqrtReal( n2 );
				}

				//Curvature @(j-1/2,i-1/2) = (J,I)
				nx = ctx.nv.x(j,i);
				ny = ctx.nv.y(j,i);
				n2 = PetscSqr(nx) + PetscSqr(ny);
				if (n2 > epsilon)
				{
					kappa += ( - nx - ny ) / PetscSqrtReal( n2 );
				}

				//Curvature @(j,i)
				kappa *= 0.5;

				//Surface force @(j,i)
				nx = 0.25 * ( ctx.nv.x(j+1,i+1) + ctx.nv.x(j,i+1) + ctx.nv.x(j+1,i) + ctx.nv.x(j,i) );
				ny = 0.25 * ( ctx.nv.y(j+1,i+1) + ctx.nv.y(j,i+1) + ctx.nv.y(j+1,i) + ctx.nv.y(j,i) );
				nmag = PetscSqrtReal( nx*nx + ny*ny );
				ctx.sFTmp.x(j,i) = kappa * nmag * nx;
				ctx.sFTmp.y(j,i) = kappa * nmag * ny;
			}
			RestoreFieldArray(ctx.sFTmp);
			RestoreGhostedFieldArray(ctx.nv);
		}

		//Interpolate surface force on staggered grid
		{
			GetGhostedFieldArray(ctx.sFTmp);
			ForAllJI(ctx.da)
			{
				ctx.sFx(b,j,i) = LowOrder::InterpolateEdge(ctx.sFTmp.x(),X_,j,i);
				ctx.sFy(b,j,i) = LowOrder::InterpolateEdge(ctx.sFTmp.y(),Y_,j,i);
			}
			RestoreGhostedFieldArray(ctx.sFTmp);
		}
	}
	RestoreFieldArray(ctx.sFx);
	RestoreFieldArray(ctx.sFy);
	RestoreGhostedFieldArray(ctx.phi);

	return(0);
}


PetscErrorCode PhiEquation::MakeFoam
(
	Context::Ctx &ctx
)
{
	SumVectorField(ctx.phi,ctx.phiSum);
	return(0);
}


PetscErrorCode PhiEquation::ComputeChemicalPotential
(
	Context::Ctx &ctx
)
{
	//to do
	return(0);
}


PetscErrorCode PhiEquation::FormSNESFunction
(
	SNES  snes,
	Vec   g_X,
	Vec   g_F,
	void *ptr
)
{
	//Get context
	Context::Ctx &ctx = *(Context::Ctx*)ptr;

	//Create local Petsc vector X = phi^{n+1}
	Vec l_X;
	PetscScalar **a_X;
	GetLocalPetscVector(ctx.da,l_X);
	GetGhostedPetscArray(ctx.da,g_X,l_X,a_X);

	//F
	PetscScalar **a_F;
	GetPetscArray(ctx.da,g_F,a_F);

	//Get RHS
	GetFieldArray(ctx.phiRHS,ctx.b);

	//Loop
	ForAllJI(ctx.da)
	{
		// F = (I - theta . nabla^2) X^(n+1) - rhs(X^n)
		PetscScalar X = a_X[j][i];
		PetscScalar theta = (ctx.xi2DtInvPeImp);
		PetscScalar LapX = FourthOrder::Laplacian(a_X,j,i);
		PetscScalar RHS = ctx.phiRHS(ctx.b,j,i);
		a_F[j][i] = ( X - theta * LapX ) - RHS;
	}

	//Restore local vector X
	RestoreGhostedPetscArray(ctx.da,l_X,a_X);
	RestoreLocalPetscVector(ctx.da,l_X);

	//Restore F
	RestorePetscArray(ctx.da,g_F,a_F);

	//Restore RHS
	RestoreFieldArray(ctx.phiRHS,ctx.b);

	return(0);
}


PetscErrorCode PhiEquation::FormSNESJacobian
(
	SNES  snes,
	Vec   g_X,
	Mat   jac,
	Mat   pjac,
	void *ptr
)
{
	Context::Ctx &ctx = *(Context::Ctx*)ptr;
	MatZeroEntries(pjac);
	const PetscInt n = 9;
	MatStencil row;
	MatStencil col[n];
	PetscScalar vec[n];
	ForAllJI(ctx.da)
	{
		//LHS = I - theta . nabla^2
		PetscScalar theta = (ctx.xi2DtInvPeImp);
		FourthOrder::IMinusLaplacian(theta,j,i,row,col,vec);
		MatSetValuesStencil(pjac,1,&row,n,col,vec,ADD_VALUES);
	}
	MatAssemblyBegin(pjac,MAT_FINAL_ASSEMBLY);
	MatAssemblyEnd(pjac,MAT_FINAL_ASSEMBLY);
	return(0);
}

// ForAllB(ctx.phi.n){
// PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
// printf(" rank = %i V[%i] = %4.5f V0[%i] = %4.5f\n",rank,b,ctx.V[b],ctx.V0[b]);}
