#include "PressureEquation.h"
#include "Field2D.h"
#include "Context.h"
#include "PetscWrapper.h"
#include "Interpolation2D.h"
#include "Utils.h"


using namespace Field2D;
using namespace Interpolation2D::LowOrder;
using namespace Context;


namespace
{
	const PetscInt X_ = 0; /**< \brief Space direction = 0 */
	const PetscInt Y_ = 1; /**< \brief Space direction = 1 */
	const PetscInt C_ = 0; /**< \brief Centre index = 0 */
	const PetscInt N_ = 1; /**< \brief North index = 1 */
	const PetscInt S_ = 2; /**< \brief South index = 2 */
	const PetscInt E_ = 3; /**< \brief East index = 3 */
	const PetscInt W_ = 4; /**< \brief West index = 4 */
}


PetscErrorCode PressureEquation::FormSNESFunction
(
	SNES  snes,
	Vec   g_X,
	Vec   g_F,
	void *ptr
)
{
	///Get context
	Ctx &ctx = *(Ctx*)ptr;

	//Create local Petsc vector X = pressure^{n+1}
	Vec l_X;
	PetscScalar **a_X;
	PetscWrapper::GetLocalPetscVector(ctx.da,l_X);
	PetscWrapper::GetGhostedPetscArray(ctx.da,g_X,l_X,a_X);

	//F
	PetscScalar **a_F;
	PetscWrapper::GetPetscArray(ctx.da,g_F,a_F);

	PetscScalar invRho[5];
	GetFieldArray(ctx.divRHS);
	GetGhostedFieldArray(ctx.rho);
	ForAllJI(ctx.da)
	{
		//F = ( Div[ Grad(p) / rho ] - Div(U*) ) x (dx^2)
		InterpolateInverseDensitiesAtCNSEW(ctx.rho(),j,i,invRho);
		a_F[j][i] =
		(
			  invRho[C_] * a_X[j  ][i  ]
			+ invRho[N_] * a_X[j+1][i  ]
			+ invRho[S_] * a_X[j-1][i  ]
			+ invRho[E_] * a_X[j  ][i+1]
			+ invRho[W_] * a_X[j  ][i-1]
		) - ctx.divRHS(j,i) * (ctx.dx);
	}

	//Restore local vector X
	PetscWrapper::RestoreGhostedPetscArray(ctx.da,l_X,a_X);
	PetscWrapper::RestoreLocalPetscVector(ctx.da,l_X);

	//Restore F
	PetscWrapper::RestorePetscArray(ctx.da,g_F,a_F);

	//Restore context fields
	RestoreFieldArray(ctx.divRHS);
	RestoreGhostedFieldArray(ctx.rho);

	return(0);
}


PetscErrorCode PressureEquation::FormSNESJacobian
(
	SNES  snes,
	Vec   g_X,
	Mat   jac,
	Mat   pjac,
	void *ptr
)
{
	Ctx& ctx = *(Ctx*)ptr;
	MatZeroEntries(pjac);
	MatStencil row;
	MatStencil col[5];
	PetscScalar vec[5];
	PetscScalar invRho[5];
	GetFieldArray(ctx.divRHS);
	GetGhostedFieldArray(ctx.rho);
	ForAllJI(ctx.da)
	{
		//LHS = Div[ Grad / rho ] x (dx^2)
		InterpolateInverseDensitiesAtCNSEW(ctx.rho(),j,i,invRho);
		row.j = j;
		row.i = i;
		vec[0] = invRho[C_]; col[0].j = (j  ); col[0].i = (i  );
		vec[1] = invRho[N_]; col[1].j = (j+1); col[1].i = (i  );
		vec[2] = invRho[S_]; col[2].j = (j-1); col[2].i = (i  );
		vec[3] = invRho[E_]; col[3].j = (j  ); col[3].i = (i+1);
		vec[4] = invRho[W_]; col[4].j = (j  ); col[4].i = (i-1);
		MatSetValuesStencil(pjac,1,&row,5,col,vec,ADD_VALUES);
	}
	RestoreFieldArray(ctx.divRHS);
	RestoreGhostedFieldArray(ctx.rho);
	MatAssemblyBegin(pjac,MAT_FINAL_ASSEMBLY);
	MatAssemblyEnd(pjac,MAT_FINAL_ASSEMBLY);
	return(0);
}
