#include "UEquation.h"
#include "Constant.h"
#include "Interpolation2D.h"
#include "Discretisation2D.h"
#include "Utils.h"


using namespace Field2D;
using namespace Interpolation2D::LowOrder;
using namespace Discretisation2D;


namespace
{
	const PetscInt X_ = 0; /**< \brief Space direction = 0 */
	const PetscInt Y_ = 1; /**< \brief Space direction = 1 */
	const PetscInt C_ = 0; /**< \brief Centre index = 0 */
	const PetscInt N_ = 1; /**< \brief North index = 1 */
	const PetscInt S_ = 2; /**< \brief South index = 2 */
	const PetscInt E_ = 3; /**< \brief East index = 3 */
	const PetscInt W_ = 4; /**< \brief West index = 4 */
}


PetscErrorCode UEquation::ComputeEtaRho
(
	Context::Ctx &ctx
)
{
	const PetscScalar lambda = 0.5;
	auto SmoothHeaviside = [&lambda](const PetscScalar& x)
	{
		//Smoothed Heaviside step function
		//Minjeaud, 2013, An unconditionally stable uncoupled scheme for a triphasic Cahn-Hilliard/Navier-Stokes model
		return x < -lambda ? 0. :
		       x >  lambda ? 1. :
		       0.5 + 0.5 * ( x / lambda + PetscSinScalar( PETSC_PI * x / lambda ) / PETSC_PI );
	};
	GetFieldArray(ctx.phi);
	GetFieldArray(ctx.eta);
	GetFieldArray(ctx.rho);
	ForAllB(ctx.phi.n) //wrong ??
	{
		ForAllJI(ctx.da)
		{
			PetscScalar shiftedPhi = ctx.phi(b,j,i) - 0.5;
			PetscScalar smoothPhi = SmoothHeaviside(shiftedPhi);
			ctx.eta(j,i) = 1. + smoothPhi * ( (Constant::Fluid::EtaRatio)-1 );
			ctx.rho(j,i) = 1. + smoothPhi * ( (Constant::Fluid::RhoRatio)-1 );
		}
	}
	RestoreFieldArray(ctx.phi);
	RestoreFieldArray(ctx.eta);
	RestoreFieldArray(ctx.rho);

	//Mean density
	ComputeArithmeticMeanOfFieldElements(ctx.rho,ctx.avgRho);
	return(0);
}


PetscErrorCode UEquation::ComputeFConv
(
	Context::Ctx &ctx,
	PetscInt j,
	PetscInt i
)
{
	PetscScalar uConv[2];

	//@(j,I)
	uConv[X_] = ctx.u.x(j,i);                     //@ (j,i-1/2) = (j    ,I)
	uConv[Y_] = InterpolateFace(ctx.u.y(),j+1,i); //@ (j,i-1/2) = (J+1/2,I)
	ctx.fConv.x(j,i) = FirstOderUpwind::UGrad(uConv,ctx.u.x(),j,i);

	//@(J,i)
	uConv[X_] = InterpolateFace(ctx.u.x(),j,i+1); //@ (j-1/2,i) = (J,I+1/2)
	uConv[Y_] = ctx.u.y(j,i);                     //@ (j-1/2,i) = (J,i    )
	ctx.fConv.y(j,i) = FirstOderUpwind::UGrad(uConv,ctx.u.y(),j,i);
	return(0);
}


PetscErrorCode UEquation::ComputeFViscAndFViscT
(
	Context::Ctx &ctx,
	const PetscScalar *const invRho,
	PetscInt j,
	PetscInt i
)
{
	PetscScalar etas[5];
	ForAllDim()
	{
		InterpolateViscositiesAtCNSEW(ctx.eta(),dim,j,i,etas);
		ctx.fVisc(dim,j,i) = invRho[dim] * ( SecondOrder::Laplacian(etas,ctx.u(dim),j,i) );
		ctx.fViscT(dim,j,i) = invRho[dim] * ( SecondOrder::TransposedLaplacian(etas,ctx.u(),dim,j,i) );
	}
	return(0);
}


PetscErrorCode UEquation::ComputeFCap
(
	Context::Ctx &ctx,
	const PetscScalar *const invRho,
	PetscInt j,
	PetscInt i
)
{
	ctx.fCap.x(j,i) = 0;
	ctx.fCap.y(j,i) = 0;
	ForAllB(ctx.phi.n)
	{
		ctx.fCap.x(j,i) += ctx.sFx(b,j,i);
		ctx.fCap.y(j,i) += ctx.sFy(b,j,i);
	}
	return(0);
}


PetscErrorCode UEquation::ComputeFExt
(
	Context::Ctx &ctx,
	const PetscScalar *const invRho,
	PetscInt j,
	PetscInt i
)
{
	ForAllDim()
	{
		ctx.fExt(dim,j,i) = ( 1. - ctx.avgRho * invRho[dim] ) * ( Constant::External::Gravity[dim] );
	}
	return(0);
}


PetscErrorCode UEquation::ComputeRHS
(
	Context::Ctx &ctx
)
{
	GetFieldArray(ctx.uRHS);
	GetFieldArray(ctx.fConv);
	GetFieldArray(ctx.fVisc);
	GetFieldArray(ctx.fViscT);
	GetFieldArray(ctx.fCap);
	GetFieldArray(ctx.fExt);
	GetFieldArray(ctx.sFx);
	GetFieldArray(ctx.sFy);
	GetGhostedFieldArray(ctx.u);
	GetGhostedFieldArray(ctx.eta);
	GetGhostedFieldArray(ctx.rho);
	ForAllJI(ctx.da)
	{
// 		UEquation::ComputeFConv(ctx,j,i);
		PetscScalar invRho[2] =
		{
			InterpolateEdgeInverse(ctx.rho(),X_,j,i),
			InterpolateEdgeInverse(ctx.rho(),Y_,j,i)
		};
// 		UEquation::ComputeFViscAndFViscT(ctx,invRho,j,i);
// 		UEquation::ComputeFCap(ctx,invRho,j,i);
		UEquation::ComputeFExt(ctx,invRho,j,i);
		ForAllDim()
		{
			ctx.uRHS(dim,j,i) =
				  ctx.u(dim,j,i)
				- ctx.fConv (dim,j,i) * (ctx.dtInvDx1)
				+ ctx.fVisc (dim,j,i) * (ctx.dtInvDx2InvRe)
				+ ctx.fViscT(dim,j,i) * (ctx.dtInvDx2InvRe)
				- ctx.fCap  (dim,j,i) * (ctx.dtXiInvDx3InvWe)
				+ ctx.fExt  (dim,j,i) * (ctx.dt)
			;
		}
	}
	RestoreFieldArray(ctx.uRHS);
	RestoreFieldArray(ctx.fConv);
	RestoreFieldArray(ctx.fVisc);
	RestoreFieldArray(ctx.fViscT);
	RestoreFieldArray(ctx.fCap);
	RestoreFieldArray(ctx.fExt);
	RestoreFieldArray(ctx.sFx);
	RestoreFieldArray(ctx.sFy);
	RestoreGhostedFieldArray(ctx.u);
	RestoreGhostedFieldArray(ctx.eta);
	RestoreGhostedFieldArray(ctx.rho);
	return(0);
}


PetscErrorCode UEquation::ComputeDivRHS
(
	Context::Ctx &ctx
)
{
	GetFieldArray(ctx.divRHS);
// 	GetGhostedFieldArray(ctx.u);
	GetGhostedFieldArray(ctx.fConv);
	GetGhostedFieldArray(ctx.fVisc);
	GetGhostedFieldArray(ctx.fViscT);
	GetGhostedFieldArray(ctx.fCap);
	GetGhostedFieldArray(ctx.fExt);
	ForAllJI(ctx.da)
	{
		ctx.divRHS(j,i) =
		(
// 			  SecondOrder::Divergent(ctx.u(),j,i) //not necessary since div(u^n) = 0
			- SecondOrder::Divergent(ctx.fConv() ,j,i) * (ctx.dtInvDx1)
			+ SecondOrder::Divergent(ctx.fVisc() ,j,i) * (ctx.dtInvDx2InvRe)
			+ SecondOrder::Divergent(ctx.fViscT(),j,i) * (ctx.dtInvDx2InvRe)
			- SecondOrder::Divergent(ctx.fCap()  ,j,i) * (ctx.dtXiInvDx3InvWe)
			+ SecondOrder::Divergent(ctx.fExt()  ,j,i) * (ctx.dt)
		); //(1/dx) moved to pressure equation
	}
	RestoreFieldArray(ctx.divRHS);
// 	RestoreGhostedFieldArray(ctx.u   );
	RestoreGhostedFieldArray(ctx.fConv);
	RestoreGhostedFieldArray(ctx.fVisc);
	RestoreGhostedFieldArray(ctx.fViscT);
	RestoreGhostedFieldArray(ctx.fCap);
	RestoreGhostedFieldArray(ctx.fExt);
	return (0);
}


PetscErrorCode UEquation::SubstractGradP
(
	Context::Ctx &ctx
)
{
	GetFieldArray(ctx.uRHS);
	GetGhostedFieldArray(ctx.rho);
	GetGhostedFieldArray(ctx.p);
	ForAllJI(ctx.da)
	{
		ForAllDim()
		{
			PetscScalar invRho = InterpolateEdgeInverse(ctx.rho(),dim,j,i);
			PetscScalar gradP = SecondOrder::Grad(ctx.p(),dim,j,i);
			ctx.uRHS(dim,j,i) -= invRho * gradP * (ctx.invDx1);
		}
	}
	RestoreFieldArray(ctx.uRHS);
	RestoreGhostedFieldArray(ctx.rho);
	RestoreGhostedFieldArray(ctx.p);
	return (0);
}


PetscErrorCode UEquation::SubstractMomentumExcess
(
	Context::Ctx &ctx
)
{
	PetscScalar localRhoU[2] = {0,0}, localRho[2] = {0,0};
	PetscScalar globalRhoU[2], globalRho[2];
	GetFieldArray(ctx.u);
	GetGhostedFieldArray(ctx.rho);
	{
		ForAllJI(ctx.da)
		{
			ForAllDim()
			{
				PetscScalar rhoInterp = InterpolateEdge(ctx.rho(),dim,j,i);
				localRhoU[dim]+= rhoInterp * ctx.u(dim,j,i);
				localRho[dim]+= rhoInterp;
			}
		}
	}
	MPI_Allreduce(localRhoU,globalRhoU,2,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD); //MPIU_Allreduce ?
	MPI_Allreduce(localRho,globalRho,2,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
	{
		ForAllJI(ctx.da)
		{
			ForAllDim()
			{
				ctx.u(dim,j,i) -= globalRhoU[dim]/globalRho[dim];
			}
		}
	}
	RestoreFieldArray(ctx.u);
	RestoreGhostedFieldArray(ctx.rho);
	return (0);
}
